import pytest
import os
from paraphase.prepare_bam_and_vcf import VcfGenerater


class TestVcfGenerater(object):
    cur_dir = os.path.dirname(__file__)
    sample_dir = os.path.join(cur_dir, "test_data")
    sample_id = "HG00733"
    vcf_generater = VcfGenerater(
        sample_id,
        sample_dir,
        None,
    )

    def test_get_var(self):
        all_bases = ["A"] * 10
        var_seq, dp, ad, gt, qual = VcfGenerater.get_var(all_bases, "A")
        assert gt == "0"
        assert var_seq == "A"
        assert dp == 10
        assert ad == 10

        all_bases = []
        var_seq, dp, ad, gt, qual = VcfGenerater.get_var(all_bases, "A")
        assert gt == "."
        assert var_seq == "A"
        assert dp == 0
        assert ad == 0

        all_bases = ["T"] * 2
        var_seq, dp, ad, gt, qual = VcfGenerater.get_var(all_bases, "A")
        assert gt == "1"
        assert var_seq == "T"
        assert dp == 2
        assert ad == 2

        all_bases = ["*"] * 3
        var_seq, dp, ad, gt, qual = VcfGenerater.get_var(all_bases, "A")
        assert gt == "1"
        assert var_seq == "*"
        assert dp == 3
        assert ad == 3

        all_bases = ["T", "T", "A"]
        var_seq, dp, ad, gt, qual = VcfGenerater.get_var(all_bases, "A")
        assert gt == "1"
        assert var_seq == "T"
        assert dp == 3
        assert ad == 2

        all_bases = ["A+2T", "A+2T", "A"]
        var_seq, dp, ad, gt, qual = VcfGenerater.get_var(all_bases, "A")
        assert gt == "1"
        assert var_seq == "A+2T"
        assert dp == 3
        assert ad == 2

        all_bases = ["A+2T", "A+2T", "A+2T", "A+2T", "A"]
        var_seq, dp, ad, gt, qual = VcfGenerater.get_var(all_bases, "A")
        assert gt == "1"
        assert var_seq == "A+2T"
        assert dp == 5
        assert ad == 4



import pytest
import os
from paraphase.genes.rccx_phaser import RccxPhaser


class TestRccxPhaser(object):

    cur_dir = os.path.dirname(__file__)
    sample_dir = os.path.join(cur_dir, "test_data")
    sample_id = "HG00733"
    phaser = RccxPhaser(sample_id, sample_dir)

    def test_annotate_var(self):
        allele_var = [[], []]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "WT"

        allele_var = [[], ["var1"]]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "WT"

        allele_var = [["var1"], ["var1", "var2"]]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "var1"

        allele_var = [["var1", "var2", "var3"], ["var1", "var2"]]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "var1,var2"

        allele_var = [[]]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "pseudogene_deletion"

        allele_var = [["var1"]]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "deletion_var1"

        allele_var = [["var1", "var2"], [], []]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "gene_duplication"

        allele_var = [
            ["var1", "var2", "var3", "var4", "var5", "var6"],
            ["var1", "var2", "var3", "var4", "var5", "var6"],
            [],
        ]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "pseudogene_duplication"

        allele_var = [["var1", "var2"], ["var1"], []]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "duplication_WT_plus_var1"

        allele_var = [
            ["var1", "var2", "var3", "var4", "var5", "var6"],
            ["var1", "var2", "var3", "var4", "var5", "var6"],
            ["var1", "var2"],
        ]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "var1,var2_pseudogene_duplication"

        allele_var = [
            ["var1", "var2", "var3", "var4", "var5", "var6"],
            ["var1", "var2", "var3", "var4"],
            ["var1", "var2"],
        ]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele == "duplication_var1,var2_plus_var1,var2,var3,var4"

        allele_var = [
            [],
            [],
            [],
            ["var1", "var2"],
        ]
        annotated_allele = self.phaser.annotate_var(allele_var)
        assert annotated_allele is None

    def test_annotate_alleles(self):
        ending_copies = []
        ass_haps = ["hap1", "hap2", "hap3", "hap4"]
        two_cp_haplotypes = []
        alleles = [["hap1", "hap2"], ["hap3", "hap4"]]
        hap_variants = {
            "hap1": [],
            "hap2": [],
            "hap3": [],
            "hap4": [],
        }
        annotated_alleles = self.phaser.annotate_alleles(
            True,
            alleles,
            hap_variants,
            ending_copies,
            ass_haps,
            two_cp_haplotypes,
        )
        assert annotated_alleles == ["WT", "WT"]

        annotated_alleles = self.phaser.annotate_alleles(
            False,
            alleles,
            hap_variants,
            ending_copies,
            ass_haps,
            two_cp_haplotypes,
        )
        assert annotated_alleles == []

        ending_copies = ["hap1", "hap2"]
        annotated_alleles = self.phaser.annotate_alleles(
            False,
            alleles,
            hap_variants,
            ending_copies,
            ass_haps,
            two_cp_haplotypes,
        )
        assert annotated_alleles == ["WT", "WT"]

        hap_variants = {
            "hap1": ["var1"],
            "hap2": [],
            "hap3": [],
            "hap4": [],
        }
        ending_copies = ["hap1", "hap2"]
        annotated_alleles = self.phaser.annotate_alleles(
            False,
            alleles,
            hap_variants,
            ending_copies,
            ass_haps,
            two_cp_haplotypes,
        )
        assert annotated_alleles == ["var1", "WT"]



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


from collections import namedtuple
from ..phaser import Phaser


class Opn1lwPhaser(Phaser):
    fields = (
        "total_cn",
        "opn1lw_cn",
        "opn1mw_cn",
        "final_haplotypes",
        "two_copy_haplotypes",
        "first_copies",
        "last_copies",
        "middle_copies",
        "annotated_haplotypes",
        "phasing_success",
        "annotated_alleles",
        "alleles_final",
        "alleles_full",
        "hap_links",
        "directional_links",
        "links_loose",
        "alleles_raw",
        "highest_total_cn",
        "assembled_haplotypes",
        "sites_for_phasing",
        "unique_supporting_reads",
        "het_sites_not_used_in_phasing",
        "homozygous_sites",
        "haplotype_details",
        "nonunique_supporting_reads",
        "read_details",
        "genome_depth",
        "region_depth",
        "sample_sex",
    )
    GeneCall = namedtuple(
        "GeneCall",
        fields,
        defaults=(None,) * len(fields),
    )
    pathogenic_haps = ["LIAVA", "LVAVA", "LIAVS", "MIAVA", "MVVVA", "MVAVA", "LIAIA"]

    def __init__(
        self, sample_id, outdir, genome_depth=None, genome_bam=None, sample_sex=None
    ):
        Phaser.__init__(self, sample_id, outdir, genome_depth, genome_bam, sample_sex)

    def set_parameter(self, config):
        super().set_parameter(config)
        self.exon3_variants = config.get("exon3_variants")
        self.pivot_vars = config.get("pivot_vars")
        self.last_copy_vars = config.get("last_copy_vars")

    def call_exon3(self, vars):
        """Call known variants in Exon 3"""
        annotated_vars = [None] * 5
        i = 0
        for pos_item in self.exon3_variants:
            query_vars, alt, ref = pos_item
            if len(set(vars).intersection(set(query_vars))) == len(query_vars):
                annotated_vars[i] = alt
            else:
                annotated_vars[i] = ref
            i += 1
        if None not in annotated_vars:
            return "".join(annotated_vars)
        return None

    @staticmethod
    def phase_single_allele(
        first_copy,
        all_haps_on_this_allele,
        last_copies,
        middle_copies,
        hap_links,
        alleles,
        directional_links,
        directional_links_loose,
    ):
        """
        Find the second copy on one allele where the first copy is known.
        """
        hap = first_copy
        remaining_haps = [a for a in all_haps_on_this_allele if a != hap]
        middle_copies = [a for a in middle_copies if a in all_haps_on_this_allele]
        # two copies
        if len(remaining_haps) == 1:
            return [hap, remaining_haps[0]]

        # three copies, last one known
        remaining_noending_haps = [a for a in remaining_haps if a not in last_copies]
        if len(remaining_noending_haps) == 1:
            return [hap, remaining_noending_haps[0]]
        # three total copies. we don't know the last copy but we know the middle copy
        if len(remaining_haps) == 2 and len(middle_copies) == 1:
            return [hap, middle_copies[0]]

        # four copies and we know the one before the last copy
        if len(last_copies) == 1:
            last_copy_before = hap_links.get(last_copies[0])
            if last_copy_before is not None and len(last_copy_before) == 1:
                remaining_noending_not_before_ending_haps = [
                    a for a in remaining_noending_haps if a not in last_copy_before
                ]
                if len(remaining_noending_not_before_ending_haps) == 1:
                    return [
                        hap,
                        remaining_noending_not_before_ending_haps[0],
                    ]
        # four total copies. we know the directional link between the middle two copies
        if len(remaining_noending_haps) == 2:
            hap1, hap2 = remaining_noending_haps
            if (
                f"{hap1}-{hap2}" in directional_links
                and f"{hap2}-{hap1}" not in directional_links
            ):
                return [hap, hap1]
            if (
                f"{hap2}-{hap1}" in directional_links
                and f"{hap1}-{hap2}" not in directional_links
            ):
                return [hap, hap2]

        # if other haps are phased on this allele
        if len(alleles) == 1 and last_copies != [] and last_copies[0] in alleles[0]:
            remaining_haps_not_phased = [
                a for a in remaining_noending_haps if a not in alleles[0]
            ]
            if len(remaining_haps_not_phased) == 1:
                return [
                    hap,
                    remaining_haps_not_phased[0],
                ]
            if len(remaining_haps_not_phased) == 2:
                hap1, hap2 = remaining_haps_not_phased
                if (
                    f"{hap1}-{hap2}" in directional_links
                    and f"{hap2}-{hap1}" not in directional_links
                ):
                    return [hap, hap1]
                if (
                    f"{hap2}-{hap1}" in directional_links
                    and f"{hap1}-{hap2}" not in directional_links
                ):
                    return [hap, hap2]

        # loose links of the first copy
        hap_next_loose = directional_links_loose.get(hap)
        if hap_next_loose is not None:
            hap_next_loose_noending = [
                a
                for a in hap_next_loose
                if a not in last_copies and hap_next_loose[a] >= 3
            ]
            if (
                len(hap_next_loose_noending) == 1
                and hap_next_loose_noending[0] in all_haps_on_this_allele
            ):
                return [hap, hap_next_loose_noending[0]]

        return None

    def call(self):
        if self.check_coverage_before_analysis() is False:
            return self.GeneCall()
        self.get_homopolymer()
        self.get_candidate_pos(min_vaf=0.08)
        self.het_sites = sorted(list(self.candidate_pos))
        self.remove_noisy_sites()
        homo_sites_to_add = self.add_homo_sites()
        raw_read_haps = self.get_haplotypes_from_reads(
            check_clip=True, kept_sites=homo_sites_to_add, add_sites=self.add_sites
        )

        (
            ass_haps,
            original_haps,
            hcn,
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            read_counts,
        ) = self.phase_haps(raw_read_haps)

        tmp = {}
        for i, hap in enumerate(ass_haps):
            tmp.setdefault(hap, f"hap{i+1}")
        ass_haps = tmp

        haplotypes = None
        two_cp_haps = []
        alleles = []
        alleles = []
        hap_links = {}
        directional_links = {}
        directional_links_loose = {}
        alleles_1st_2nd = []
        first_copies = []
        last_copies = []
        middle_copies = []
        counter_lw = 0
        counter_mw = 0
        counter_unknown = 0
        annotated_alleles = []
        annotated_haps = {}
        phasing_success = False
        pivot_vars = set(self.pivot_vars)
        last_copy_vars = set(self.last_copy_vars)

        if self.het_sites != []:
            haplotypes = self.output_variants_in_haplotypes(
                ass_haps,
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
            )

            hap_rename = {}
            for hap, hap_name in ass_haps.items():
                var = set(haplotypes[hap_name]["variants"])
                var_intersect = var.intersection(pivot_vars)
                if var_intersect == set():
                    counter_lw += 1
                    renamed_hap = f"opn1lw_hap{counter_lw}"
                elif len(var_intersect) == 2:
                    counter_mw += 1
                    renamed_hap = f"opn1mw_hap{counter_mw}"
                else:
                    counter_unknown += 1
                    renamed_hap = f"opnunknown_hap{counter_unknown}"
                hap_rename.setdefault(hap_name, renamed_hap)
                # first, middle, last copies
                if hap[0] not in ["x", "0"]:
                    first_copies.append(renamed_hap)
                if var.intersection(last_copy_vars) != set():
                    last_copies.append(renamed_hap)
                elif renamed_hap not in first_copies and "x" not in hap:
                    middle_copies.append(renamed_hap)
                # annotate exon3
                hap_annotated = self.call_exon3(var)
                gene_annotated = renamed_hap.split("_")[0]
                if hap_annotated in self.pathogenic_haps:
                    gene_annotated += "_" + hap_annotated
                annotated_haps.setdefault(renamed_hap, gene_annotated)

            tmp = {}
            for hap, hap_name in ass_haps.items():
                tmp.setdefault(hap, hap_rename[hap_name])
            ass_haps = tmp

            tmp = {}
            for hap_name, hap_info in haplotypes.items():
                tmp.setdefault(hap_rename[hap_name], hap_info)
            haplotypes = tmp

            # do not look for two-copy haplotypes for now
            total_cn = len(ass_haps) + len(two_cp_haps)

            (
                alleles,
                hap_links,
                directional_links,
                directional_links_loose,
            ) = self.phase_alleles(
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
                raw_read_haps,
                ass_haps,
                reverse=self.is_reverse,
            )

            # incorrect phasing suggests haplotypes with cn > 1
            if (
                self.sample_sex is not None
                and self.sample_sex == "female"
                and len(alleles) == 1
                and sorted(alleles[0]) == sorted(ass_haps.values())
            ):
                alleles = []
                total_cn = None

            # focus on first and second copies
            if self.sample_sex is not None:
                if (self.sample_sex == "male" and len(first_copies) == 1) or (
                    self.sample_sex == "female" and len(first_copies) == 2
                ):
                    for hap in first_copies:
                        hap_next = hap_links.get(hap)
                        if hap_next is not None and len(hap_next) == 1:
                            alleles_1st_2nd.append([hap, hap_next[0]])
                        elif hap_next is None and hap in last_copies:
                            alleles_1st_2nd.append([hap])
                        else:
                            hap_next_loose = directional_links_loose.get(hap)
                            if hap_next_loose is not None and len(hap_next_loose) == 1:
                                next_hap_name, read_count = list(
                                    hap_next_loose.items()
                                )[0]
                                if read_count >= 3:
                                    alleles_1st_2nd.append([hap, next_hap_name])

                    # easier phasing for males as there is one allele
                    if self.sample_sex == "male" and alleles_1st_2nd == []:
                        allele_found = self.phase_single_allele(
                            first_copies[0],
                            list(ass_haps.values()),
                            last_copies,
                            middle_copies,
                            hap_links,
                            alleles,
                            directional_links,
                            directional_links_loose,
                        )
                        if allele_found is not None:
                            alleles_1st_2nd.append(allele_found)
                    # for females, if one allele is completely phased and the other is not phased yet
                    if self.sample_sex == "female" and len(alleles_1st_2nd) == 1:
                        complete_allele = []
                        incomplete_allele = []
                        for each_allele in alleles:
                            if (
                                set(each_allele).intersection(set(first_copies))
                                != set()
                                and set(each_allele).intersection(set(last_copies))
                                != set()
                            ):
                                complete_allele.append(each_allele)
                            else:
                                incomplete_allele.append(each_allele)
                        if len(complete_allele) == 1:
                            other_allele_copies = [
                                a
                                for a in ass_haps.values()
                                if a not in complete_allele[0]
                            ]
                            other_first_copy = [
                                a for a in first_copies if a not in complete_allele[0]
                            ]
                            other_last_copy = [
                                a for a in last_copies if a not in complete_allele[0]
                            ]
                            if len(other_first_copy) == 1 and len(other_last_copy) == 1:
                                allele_found = self.phase_single_allele(
                                    other_first_copy[0],
                                    other_allele_copies,
                                    other_last_copy,
                                    middle_copies,
                                    hap_links,
                                    incomplete_allele,
                                    directional_links,
                                    directional_links_loose,
                                )
                                if allele_found is not None:
                                    alleles_1st_2nd.append(allele_found)

                    if (self.sample_sex == "male" and len(alleles_1st_2nd) == 1) or (
                        self.sample_sex == "female" and len(alleles_1st_2nd) == 2
                    ):
                        # phasing success is defined as the first two copies being phased
                        # on each allele
                        phasing_success = True
                    else:
                        for hap in first_copies:
                            if True not in [hap in a for a in alleles_1st_2nd]:
                                alleles_1st_2nd.append([hap, None])

                # annotate each pair of 1st and 2nd copies
                for each_allele in alleles_1st_2nd:
                    each_allele_annotated = []
                    for each_hap in each_allele:
                        if each_hap is not None:
                            hap_vars = haplotypes[each_hap]["variants"]
                            hap_annotated = self.call_exon3(hap_vars)
                            gene_annotated = each_hap.split("_")[0]
                            if hap_annotated in self.pathogenic_haps:
                                gene_annotated += "_" + hap_annotated
                        else:
                            gene_annotated = None
                        each_allele_annotated.append(gene_annotated)
                    annotated_alleles.append(each_allele_annotated)

        # homozygous cases
        elif self.sample_sex is not None:
            baseline_cn = None
            phasing_success = True
            if self.sample_sex == "female":
                baseline_cn = 2
            elif self.sample_sex == "male":
                baseline_cn = 1
            hap_annotated = self.call_exon3(self.homo_sites)
            if set(self.homo_sites).intersection(pivot_vars) != set():
                total_cn = baseline_cn
                counter_mw = baseline_cn
                gene_annotated = "opn1mw"
            else:
                total_cn = baseline_cn
                counter_lw = baseline_cn
                gene_annotated = "opn1lw"
            if hap_annotated in self.pathogenic_haps:
                gene_annotated += "_" + hap_annotated
            annotated_alleles.append([gene_annotated])
            if self.sample_sex == "female":
                annotated_alleles.append([gene_annotated])

        self.close_handle()

        return self.GeneCall(
            total_cn,
            counter_lw,
            counter_mw,
            ass_haps,
            two_cp_haps,
            first_copies,
            last_copies,
            middle_copies,
            annotated_haps,
            phasing_success,
            annotated_alleles,
            alleles_1st_2nd,
            alleles,
            hap_links,
            directional_links,
            directional_links_loose,
            alleles,
            hcn,
            original_haps,
            self.het_sites,
            uniquely_supporting_reads,
            self.het_no_phasing,
            self.homo_sites,
            haplotypes,
            nonuniquely_supporting_reads,
            raw_read_haps,
            self.mdepth,
            self.region_avg_depth._asdict(),
            self.sample_sex,
        )



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


from ..phaser import Phaser


class Cfc1Phaser(Phaser):
    def __init__(
        self, sample_id, outdir, genome_depth=None, genome_bam=None, sample_sex=None
    ):
        Phaser.__init__(self, sample_id, outdir, genome_depth, genome_bam, sample_sex)

    def call(self):
        if self.check_coverage_before_analysis() is False:
            return self.GeneCall()
        self.get_homopolymer()
        self.get_candidate_pos()
        self.het_sites = sorted(list(self.candidate_pos))
        self.remove_noisy_sites()
        homo_sites_to_add = self.add_homo_sites()
        raw_read_haps = self.get_haplotypes_from_reads(
            kept_sites=homo_sites_to_add, add_sites=self.add_sites
        )

        (
            ass_haps,
            original_haps,
            hcn,
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            read_counts,
        ) = self.phase_haps(raw_read_haps)

        tmp = {}
        for i, hap in enumerate(ass_haps):
            tmp.setdefault(hap, f"hap{i+1}")
        ass_haps = tmp

        haplotypes = None
        if self.het_sites != []:
            haplotypes = self.output_variants_in_haplotypes(
                ass_haps,
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
            )

        two_cp_haps = []
        if len(ass_haps) == 2:
            two_cp_haps = list(ass_haps.values())
        elif len(ass_haps) == 3:
            two_cp_haps = self.compare_depth(haplotypes, loose=True)
            if two_cp_haps == [] and read_counts is not None:
                # check if one smn1 haplotype has more reads than others
                haps = list(read_counts.keys())
                counts = list(read_counts.values())
                max_count = max(counts)
                cp2_hap = haps[counts.index(max_count)]
                others_max = sorted(counts, reverse=True)[1]
                probs = self.depth_prob(max_count, others_max)
                if probs[0] < 0.15 and others_max >= 10:
                    two_cp_haps.append(ass_haps[cp2_hap])

        total_cn = len(ass_haps) + len(two_cp_haps)
        if total_cn < 4:
            total_cn = None

        self.close_handle()

        return self.GeneCall(
            total_cn,
            None,
            ass_haps,
            two_cp_haps,
            None,
            None,
            hcn,
            original_haps,
            self.het_sites,
            uniquely_supporting_reads,
            self.het_no_phasing,
            self.homo_sites,
            haplotypes,
            nonuniquely_supporting_reads,
            raw_read_haps,
            self.mdepth,
            self.region_avg_depth._asdict(),
            self.sample_sex,
        )



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


import numpy as np
import pysam


class GenomeDepth:
    def __init__(self, bam, genome_depth_region_file):
        self.bam = bam
        self.genome_depth_region_file = genome_depth_region_file
        self._bamh = pysam.AlignmentFile(bam, "rb")
        self.mdepth = None
        self.mad = None
        self.x = []
        self.y = []
        self.chr_in_name = GenomeDepth.check_chr_name(self._bamh)

    @staticmethod
    def check_chr_name(bam_handle):
        """
        Check bam header to see if chromosome names contain "chr"
        """
        header = bam_handle.header
        header = header.to_dict()
        sq = header.get("SQ")
        if sq is not None:
            chr_names = [a.get("SN") for a in sq]
            if True not in ["chr" in a for a in chr_names]:
                return False
        return True

    def get_genome_depth(self):
        depth = []
        with open(self.genome_depth_region_file) as f:
            for line in f:
                if line[0] != "#":
                    at = line.split()
                    nchr = at[0]
                    if self.chr_in_name is False:
                        nchr = nchr.replace("chr", "")
                    pos1 = int(at[1])
                    for pos in [pos1, pos1 + 1600]:
                        site_depth = self._bamh.count(
                            nchr, pos - 1, pos, read_callback="all"
                        )
                        depth.append(site_depth)
        self.mdepth = np.median(depth)
        self.mad = np.median([abs(a - self.mdepth) for a in depth]) / self.mdepth

    def call(self):
        self.get_genome_depth()
        self._bamh.close()
        return self.mdepth, self.mad

    def check_sex(self):
        """Determine sample sex based on coverage"""
        with open(self.genome_depth_region_file) as f:
            for line in f:
                if line[0] != "#":
                    at = line.split()
                    nchr = at[0]
                    if self.chr_in_name is False:
                        nchr = nchr.replace("chr", "")
                    pos1 = int(at[1])
                    site_depth = self._bamh.count(
                        nchr, pos1 - 1, pos1, read_callback="all"
                    )
                    if "X" in nchr:
                        self.x.append(site_depth)
                    elif "Y" in nchr:
                        self.y.append(site_depth)
        cov_x, cov_y = np.median(self.x), np.median(self.y)
        self._bamh.close()
        if cov_y / cov_x < 0.05:
            return "female"
        elif cov_y / cov_x > 0.1:
            if cov_x > 1.95 * cov_y:
                return "female"
            return "male"
        return None



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>

import copy
from collections import namedtuple
import pysam
from ..phaser import Phaser


class HbaPhaser(Phaser):
    new_fields = copy.deepcopy(Phaser.fields)
    new_fields.remove("gene_cn")
    new_fields.insert(3, "genotype")
    new_fields.insert(4, "sv_called")
    new_fields.insert(5, "surrounding_region_depth")
    GeneCall = namedtuple(
        "GeneCall",
        new_fields,
        defaults=(None,) * len(new_fields),
    )

    def __init__(
        self, sample_id, outdir, genome_depth=None, genome_bam=None, sample_sex=None
    ):
        Phaser.__init__(self, sample_id, outdir, genome_depth, genome_bam, sample_sex)

    def set_parameter(self, config):
        super().set_parameter(config)
        self.depth_region = config["depth_region"]

    def call(self):
        if self.check_coverage_before_analysis() is False:
            return self.GeneCall()
        genome_bamh = pysam.AlignmentFile(self.genome_bam, "rb")
        surrounding_region_depth = self.get_regional_depth(
            genome_bamh, self.depth_region
        )[0].median
        genome_bamh.close()

        self.get_homopolymer()
        self.get_candidate_pos()
        self.het_sites = sorted(list(self.candidate_pos))
        self.remove_noisy_sites()
        raw_read_haps = self.get_haplotypes_from_reads(check_clip=True)

        (
            ass_haps,
            original_haps,
            hcn,
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            read_counts,
        ) = self.phase_haps(raw_read_haps)

        count_hba1 = 0
        count_hba2 = 0
        count_unknown = 0
        count_del = 0
        count_dup = 0
        tmp = {}
        for i, hap in enumerate(ass_haps):
            if "x" in [hap[0], hap[-1]]:
                count_unknown += 1
                new_hap_name = f"hba_unknown_hap{count_unknown}"
            elif hap[0] == "0" and hap[-1] == "0":
                count_hba1 += 1
                new_hap_name = f"hba1_hap{count_hba1}"
            elif hap[0] == "0" and hap[-1] != "0":
                count_dup += 1
                new_hap_name = f"hba_dup_hap{count_dup}"
            elif hap[0] != "0" and hap[-1] == "0":
                count_del += 1
                new_hap_name = f"hba_del_hap{count_del}"
            else:
                count_hba2 += 1
                new_hap_name = f"hba2_hap{count_hba2}"
            tmp.setdefault(hap, new_hap_name)
        ass_haps = tmp

        haplotypes = None
        if self.het_sites != []:
            haplotypes = self.output_variants_in_haplotypes(
                ass_haps,
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
            )

        sv_called = []
        genotype = None
        for hap in ass_haps.values():
            if "del" in hap or "dup" in hap:
                sv_called.append(hap)
        two_cp_haps = []
        if count_unknown == 0 and count_del == 0:
            if count_hba1 == 1 and count_hba2 == 2:
                two_cp_haps += [a for a in ass_haps.values() if "hba1" in a]
            elif count_hba1 == 2 and count_hba2 == 1:
                two_cp_haps += [a for a in ass_haps.values() if "hba2" in a]
            elif count_hba1 == 1 and count_hba2 == 1:
                # two possible scenarios:
                # two pairs of identical copies, or big deletion on one allele
                # use depth
                if self.mdepth is not None:
                    prob = self.depth_prob(
                        int(surrounding_region_depth), self.mdepth / 2
                    )
                    # print(surrounding_region_depth, self.mdepth / 2, prob)
                    if prob[0] < 0.999:
                        two_cp_haps += [
                            a for a in ass_haps.values() if "hba1" in a or "hba2" in a
                        ]

        total_cn = len(ass_haps) + len(two_cp_haps)
        if self.het_sites == [] or total_cn <= 1:
            total_cn = None
        if count_del == 1 and total_cn == 3:
            genotype = "-a/aa"
        if count_del == 1 and total_cn == 4 and count_dup == 1:
            genotype = "-a/aaa"
        elif count_del == 2 and total_cn == 2:
            genotype = "-a/-a"
        if count_del == 0 and total_cn == 4 and count_dup == 0:
            genotype = "aa/aa"
        if count_del == 0 and total_cn == 5 and count_dup == 1:
            genotype = "aa/aaa"
        if count_del == 0 and total_cn == 6 and count_dup == 2:
            genotype = "aaa/aaa"
        if count_del == 0 and total_cn == 2 and count_dup == 0:
            genotype = "--/aa"

        # phase
        alleles = []
        hap_links = {}
        if self.to_phase is True:
            (
                alleles,
                hap_links,
                _,
                _,
            ) = self.phase_alleles(
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
                raw_read_haps,
                ass_haps,
                reverse=self.is_reverse,
            )
        self.close_handle()

        return self.GeneCall(
            total_cn,
            ass_haps,
            two_cp_haps,
            genotype,
            sv_called,
            surrounding_region_depth,
            alleles,
            hap_links,
            hcn,
            original_haps,
            self.het_sites,
            uniquely_supporting_reads,
            self.het_no_phasing,
            self.homo_sites,
            haplotypes,
            nonuniquely_supporting_reads,
            raw_read_haps,
            self.mdepth,
            self.region_avg_depth._asdict(),
            self.sample_sex,
        )



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


import pysam
import os
import copy
import numpy as np
from collections import Counter
from itertools import product
import re
import logging
from scipy.stats import poisson
from collections import namedtuple
from .haplotype_assembler import VariantGraph


class Phaser:
    clip_5p = r"^\d+S|^\d+H"
    clip_3p = r"\d+S$|\d+H$"
    deletion = r"\d+D"
    fields = [
        "total_cn",
        "gene_cn",
        "final_haplotypes",
        "two_copy_haplotypes",
        "alleles_final",
        "hap_links",
        "highest_total_cn",
        "assembled_haplotypes",
        "sites_for_phasing",
        "unique_supporting_reads",
        "het_sites_not_used_in_phasing",
        "homozygous_sites",
        "haplotype_details",
        "nonunique_supporting_reads",
        "read_details",
        "genome_depth",
        "region_depth",
        "sample_sex",
    ]
    GeneCall = namedtuple(
        "GeneCall",
        fields,
        defaults=(None,) * len(fields),
    )
    CoverageSummary = namedtuple("CoverageSummary", "median percentile80")
    MEAN_BASE_QUAL = 25

    def __init__(
        self,
        sample_id,
        outdir,
        wgs_depth=None,
        genome_bam=None,
        sample_sex=None,
    ):
        self.outdir = outdir
        self.sample_id = sample_id
        self.homopolymer_sites = {}
        self.het_sites = []  # for phasing
        self.het_no_phasing = []
        self.homo_sites = []
        self.candidate_pos = set()
        self.mdepth = wgs_depth
        self.genome_bam = genome_bam
        self.sample_sex = sample_sex

    def set_parameter(self, config):
        self.gene = config["gene"]
        self.bam = os.path.join(
            self.outdir, self.sample_id + f"_{self.gene}_realigned.bam"
        )
        if os.path.exists(self.bam) is False:
            raise Exception(f"File {self.bam} not found.")
        self._bamh = pysam.AlignmentFile(self.bam, "rb")
        self.nchr = config["nchr"]
        self.ref = config["data"]["reference"]
        self._refh = pysam.FastaFile(self.ref)
        self.pivot_site = None
        if "pivot_site" in config:
            self.pivot_site = config["pivot_site"]
        self.nchr_old = config["nchr_old"]
        self.offset = int(self.nchr_old.split("_")[1]) - 1

        # define region boundary
        self.left_boundary = config.get("left_boundary")
        self.right_boundary = config.get("right_boundary")
        if self.left_boundary is None:
            self.left_boundary = int(self.nchr_old.split("_")[1])
        if self.right_boundary is None:
            self.right_boundary = int(self.nchr_old.split("_")[2])
        self.gene_start = config.get("gene_start")
        if self.gene_start is None:
            self.gene_start = self.left_boundary
        self.gene_end = config.get("gene_end")
        if self.gene_end is None:
            self.gene_end = self.right_boundary

        self.use_supplementary = False
        if "use_supplementary" in config or "is_tandem" in config:
            self.use_supplementary = True
        self.to_phase = False
        if "to_phase" in config or "is_tandem" in config:
            self.to_phase = True
        self.is_reverse = False
        if "is_reverse" in config:
            self.is_reverse = config["is_reverse"]
        self.expect_cn2 = False
        if "expect_cn2" in config:
            self.expect_cn2 = True
        self.clip_3p_positions = []
        self.clip_5p_positions = []
        if "clip_3p_positions" in config:
            self.clip_3p_positions = config["clip_3p_positions"]
        if "clip_5p_positions" in config:
            self.clip_5p_positions = config["clip_5p_positions"]
        self.noisy_region = []
        if "noisy_region" in config:
            self.noisy_region = config["noisy_region"]
        self.add_sites = []
        if "add_sites" in config:
            self.add_sites = config["add_sites"]
        self.match = {}
        self.gene2_region = config.get("gene2_region")
        if self.gene2_region is not None:
            self.position_match = config["data"].get("gene_position_match")
            if self.position_match is not None:
                with open(self.position_match) as f:
                    for line in f:
                        at = line.split()
                        self.match.setdefault(int(at[0]), int(at[1]))

        self.del1_reads = set()
        self.del1_reads_partial = set()
        self.del2_reads = set()
        self.del2_reads_partial = set()
        self.deletion1_size = None
        self.deletion2_size = None
        self.deletion1_name = None
        self.deletion2_name = None
        self.del2_3p_pos1 = None
        self.del2_3p_pos2 = None
        self.del2_5p_pos1 = None
        self.del2_5p_pos2 = None
        self.del1_3p_pos1 = None
        self.del1_3p_pos2 = None
        self.del1_5p_pos1 = None
        self.del1_5p_pos2 = None

    def get_regional_depth(self, bam_handle, query_region, ninterval=100):
        """Get depth of the query regions"""
        region_depth = []
        for region in query_region:
            depth = []
            nstep = max(1, int((region[1] - region[0]) / ninterval))
            for pos in range(region[0], region[1], nstep):
                site_depth = bam_handle.count(
                    self.nchr, pos - 1, pos, read_callback="all"
                )
                depth.append(site_depth)
            region_depth.append(
                self.CoverageSummary(np.median(depth), np.nanpercentile(depth, 80))
            )
        return region_depth

    def check_coverage_before_analysis(self):
        """check low coverage regions for enrichment data"""
        check_region_start = self.left_boundary
        check_region_end = self.right_boundary
        if self.clip_5p_positions != []:
            check_region_start = max(self.left_boundary, max(self.clip_5p_positions))
        if self.clip_3p_positions != []:
            check_region_end = min(self.right_boundary, min(self.clip_3p_positions))
        self.region_avg_depth = self.get_regional_depth(
            self._bamh, [[check_region_start, check_region_end]]
        )[0]
        if np.isnan(self.region_avg_depth.median) or (
            self.region_avg_depth.median < 10
            and self.region_avg_depth.percentile80 < 50
        ):
            logging.warning(
                "This region does not appear to have coverage. Will not attempt to phase haplotypes."
            )
            return False
        return True

    def get_range_in_other_gene(self, pos, search_range=200):
        """
        Find the correponding coordinates in the other gene
        """
        if pos in self.match:
            return self.match[pos]
        for i in range(search_range):
            new_pos = pos + i
            if new_pos in self.match:
                return self.match[new_pos]
        return None

    def get_homopolymer(self):
        """
        Get the homopolymer and dinucleotide sites
        Variants for phasing:
        SNVs, not matching excluded bases and sites labeled as 1
        indels, not in excluded sites
        """
        seq = self._refh.fetch(self.nchr_old).upper()
        nstart = self.offset
        drepeat = {}
        # start with 5mer
        for i in range(len(seq) - 5):
            for nu in ["A", "C", "G", "T"]:
                if seq[i : i + 5].count(nu) >= 5:
                    for pos in range(i, i + 5):
                        drepeat.setdefault(pos, nu)
        # add nearby 4mers (same base), separated by stretches of a different base
        while True:
            extend = False
            for i in range(len(seq) - 4):
                for nu in ["A", "C", "G", "T"]:
                    if seq[i : i + 4].count(nu) == 4:
                        for j in range(1, 5):
                            if (
                                i + 4 + j in drepeat
                                and drepeat[i + 4 + j] == nu
                                and len(set(seq[i + 4 : i + 4 + j])) == 1
                            ):
                                for pos in range(i, i + 4 + j):
                                    if pos not in drepeat:
                                        drepeat.setdefault(pos, nu)
                                        extend = True
                            if (
                                i - 1 - j in drepeat
                                and drepeat[i - 1 - j] == nu
                                and len(set(seq[i - j : i])) == 1
                            ):
                                for pos in range(i - j, i + 4):
                                    if pos not in drepeat:
                                        drepeat.setdefault(pos, nu)
                                        extend = True
            if extend is False:
                break
        drepeat = dict(sorted(drepeat.items()))
        drepeat_keys = list(drepeat.keys())
        for i in range(len(drepeat)):
            pos = drepeat_keys[i]
            pos_adjusted = pos + nstart + 1
            base = drepeat[pos]
            if i + 1 == len(drepeat) or (
                i + 1 < len(drepeat) and drepeat_keys[i + 1] != pos + 1
            ):
                self.homopolymer_sites.setdefault(pos_adjusted, base)
                self.homopolymer_sites.setdefault(pos + 1 + nstart + 1, f"{base},1")
            elif i == 0 or (i > 0 and drepeat_keys[i - 1] != pos - 1):
                self.homopolymer_sites.setdefault(pos - 1 + nstart + 1, base)
                self.homopolymer_sites.setdefault(pos_adjusted, "A,C,G,T")
            else:
                self.homopolymer_sites.setdefault(pos_adjusted, base)

        # dinucleotide
        drepeat = {}
        # start with 4 copies of a dimer
        for i in range(len(seq) - 8):
            for dimer in [a + b for a, b in product("ACGT", repeat=2) if a != b]:
                if seq[i : i + 8].count(dimer) == 4:
                    for pos in range(i, i + 8):
                        drepeat.setdefault(pos, []).append(dimer)
        # add nearby any 2 copies of the same dimer, separated by either stretches of
        # the same 2 bases or a different dimer
        counter = 0
        while counter < 3:
            counter += 1
            extend = False
            for i in range(len(seq) - 4):
                for dimer in [a + b for a, b in product("ACGT", repeat=2) if a != b]:
                    if seq[i : i + 4].count(dimer) == 2:
                        for j in range(8):
                            pos_to_check = i + 4 + j
                            to_add = False
                            if pos_to_check in drepeat and (
                                dimer in drepeat[pos_to_check]
                                or dimer[::-1] in drepeat[pos_to_check]
                            ):
                                if j <= 1:
                                    to_add = True
                                else:
                                    mid_seq = seq[i + 4 : i + 4 + j]
                                    reduced_nu = list(set(mid_seq))
                                    if len(reduced_nu) == 2:
                                        base1 = reduced_nu[0]
                                        base2 = reduced_nu[1]
                                        if (
                                            (
                                                base1 + base1 not in mid_seq
                                                and base2 + base2 not in mid_seq
                                            )
                                            or base1 + base2 == dimer
                                            or base2 + base1 == dimer
                                        ):
                                            to_add = True
                                if to_add is True:
                                    for pos in range(i, i + 4 + j):
                                        drepeat.setdefault(pos, [])
                                        if dimer not in drepeat[pos]:
                                            drepeat[pos].append(dimer)
                                        extend = True
                            pos_to_check = i - 1 - j
                            to_add = False
                            if pos_to_check in drepeat and (
                                dimer in drepeat[pos_to_check]
                                or dimer[::-1] in drepeat[pos_to_check]
                            ):
                                if j <= 1:
                                    to_add = True
                                else:
                                    mid_seq = seq[i - j : i]
                                    reduced_nu = list(set(mid_seq))
                                    if len(reduced_nu) == 2:
                                        base1 = reduced_nu[0]
                                        base2 = reduced_nu[1]
                                        if (
                                            (
                                                base1 + base1 not in mid_seq
                                                and base2 + base2 not in mid_seq
                                            )
                                            or base1 + base2 == dimer
                                            or base2 + base1 == dimer
                                        ):
                                            to_add = True
                                if to_add is True:
                                    for pos in range(i - j, i + 4):
                                        drepeat.setdefault(pos, [])
                                        if dimer not in drepeat[pos]:
                                            drepeat[pos].append(dimer)
                                        extend = True
            if extend is False:
                break
        drepeat = dict(sorted(drepeat.items()))
        for pos in drepeat:
            pos_adjusted = pos + nstart + 1
            if pos_adjusted not in self.homopolymer_sites:
                bases = []
                for nus in drepeat[pos]:
                    bases.append(nus[0])
                    bases.append(nus[1])
                bases = list(set(bases))  # + ["1"]
                self.homopolymer_sites.setdefault(pos_adjusted, ",".join(bases))

    @staticmethod
    def depth_prob(nread, haploid_depth):
        """Find probability of cn state based on depth"""
        prob = []
        for i in range(4):
            depthexpected = (i + 1) * haploid_depth
            pmf = poisson.pmf(nread, depthexpected)
            prob.append(pmf)
        sum_prob = sum(prob)
        if sum_prob == 0:
            return None
        post_prob = [float(a) / float(sum_prob) for a in prob]
        return post_prob

    def find_big_deletion(self, min_size=5000, min_count=3, padding=50):
        """Call big deletions from data"""
        del_reads = []
        for read in self._bamh.fetch(
            self.nchr, self.left_boundary, self.right_boundary
        ):
            read_cigar = read.cigarstring
            del_len = [int(a[:-1]) for a in re.findall(Phaser.deletion, read_cigar)]
            if del_len != []:
                longest_del_size = max(del_len)
                if longest_del_size >= min_size:
                    cigar_before_del = read_cigar.split(f"{longest_del_size}D")[0]
                    del_pos = read.reference_start + sum(
                        [
                            int(a[:-1])
                            for a in re.findall(r"\d+D|\d+M", cigar_before_del)
                        ]
                    )
                    del_reads.append((del_pos, del_pos + longest_del_size))
        counter = Counter(del_reads).most_common(2)
        if len(counter) > 0 and counter[0][1] >= min_count:
            nstart, nend = counter[0][0]
            self.deletion1_size = nend - nstart
            self.deletion1_name = f"{nstart}_del_{nend-nstart}"
            self.del1_3p_pos1 = nstart - padding
            self.del1_3p_pos2 = nstart + padding
            self.del1_5p_pos1 = nend - padding
            self.del1_5p_pos2 = nend + padding
        if len(counter) > 1 and counter[1][1] >= min_count:
            nstart, nend = counter[1][0]
            self.deletion2_size = nend - nstart
            self.deletion2_name = f"{nstart}_del_{nend-nstart}"
            self.del2_3p_pos1 = nstart - padding
            self.del2_3p_pos2 = nstart + padding
            self.del2_5p_pos1 = nend - padding
            self.del2_5p_pos2 = nend + padding

    @staticmethod
    def check_del(read, del_size):
        """Find reads having the 6.3kb deletion in its cigar string"""
        del_len = [int(a[:-1]) for a in re.findall(Phaser.deletion, read.cigarstring)]
        if del_len != [] and abs(max(del_len) - del_size) < 50:
            return True
        return False

    def get_long_del_reads(
        self,
        p3_pos1,
        p3_pos2,
        p5_pos1,
        p5_pos2,
        del_size,
        min_clip_len=300,
        min_extend=1000,
    ):
        """
        Find reads having big deletions. (Improve to general SVs for future)
        Could be softclipped at either side or have the deletion in cigar.
        Parameters:
            min_clip_len (int): minimum length for the soft-clip
        Returns: fully-spanning reads (set), partially spanning reads (set)
        """
        bamh = self._bamh
        p5_reads = set()
        p3_reads = set()
        del_reads = set()
        # 3 prime clip
        pos1 = p3_pos1
        pos2 = p3_pos2
        reference_start_cutoff = pos1 - min_extend
        for read in bamh.fetch(self.nchr, pos1, pos2):
            read_name = self.get_read_name(read)
            find_clip_3p = re.findall(self.clip_3p, read.cigarstring)
            if find_clip_3p != [] and pos1 < read.reference_end < pos2:
                if (
                    int(find_clip_3p[0][:-1]) >= min_clip_len
                    and read.reference_start < reference_start_cutoff
                ):
                    p3_reads.add(read_name)
            if self.check_del(read, del_size):
                del_reads.add(read_name)
        # 5 prime clip
        pos1 = p5_pos1
        pos2 = p5_pos2
        reference_end_cutoff = pos2 + min_extend
        for read in bamh.fetch(self.nchr, pos1, pos2):
            read_name = self.get_read_name(read)
            find_clip_5p = re.findall(self.clip_5p, read.cigarstring)
            if find_clip_5p != [] and pos1 < read.reference_start < pos2:
                if (
                    int(find_clip_5p[0][:-1]) >= min_clip_len
                    and read.reference_end > reference_end_cutoff
                ):
                    p5_reads.add(read_name)
            if self.check_del(read, del_size):
                del_reads.add(read_name)
        if del_reads != set() or (p3_reads != set() and p5_reads != set()):
            return (
                del_reads.union(p3_reads.intersection(p5_reads)),
                del_reads.union(p3_reads).union(p5_reads),
            )
        return set(), set()

    def get_pivot_site_index(self):
        """Return the index of the pivot site in list of het sites"""
        positions = [int(site.split("_")[0]) for site in self.het_sites]
        if self.pivot_site in positions:
            return positions.index(self.pivot_site), True
        return -1, False

    def get_read_name(self, read):
        """Rename reads when supplementary"""
        read_name = read.query_name
        if read.is_supplementary and self.use_supplementary:
            read_name = (
                read_name + f"_sup_{read.reference_start}_{read.reference_length}"
            )
        return read_name

    def get_read_names(self, read, partial_deletion_reads):
        """Add read names for supplementary alignments"""
        read_names = [read.query_name]
        if read.is_supplementary and self.use_supplementary:
            sup_name = (
                read.query_name + f"_sup_{read.reference_start}_{read.reference_length}"
            )
            read_names = [sup_name]
            if (
                sup_name in partial_deletion_reads
                and read.query_name in partial_deletion_reads
            ):
                read_names.append(read.query_name)
        return read_names

    def get_haplotypes_from_reads(
        self,
        exclude_reads=[],
        min_mapq=5,
        min_clip_len=50,
        check_clip=False,
        partial_deletion_reads=[],
        kept_sites=[],
        add_sites=[],
        multi_allelic_sites={},
    ):
        """
        Go through reads and get bases at sites of interest.
        Two rounds, with variant site filtering in between.
        """
        raw_read_haps = self.get_haplotypes_from_reads_step(
            exclude_reads,
            min_mapq,
            min_clip_len,
            check_clip,
            partial_deletion_reads,
            multi_allelic_sites,
        )
        self.remove_var(raw_read_haps, kept_sites)
        if self.het_sites != []:
            for var in add_sites:
                if var not in self.het_sites:
                    self.het_sites.append(var)
        # add sites before 5p clip and after 3p clip
        if self.clip_5p_positions != [] and self.het_sites != []:
            clip_pos = min(self.clip_5p_positions)
            var_before_clip = [
                a for a in self.het_sites if int(a.split("_")[0]) < clip_pos
            ]
            if var_before_clip == []:
                var_pos = clip_pos - 30
                ref_base = self._refh.fetch(
                    self.nchr_old, var_pos - self.offset - 1, var_pos - self.offset
                ).upper()
                var_base = [a for a in ["A", "C", "G", "T"] if a != ref_base][0]
                new_var = f"{var_pos}_{ref_base}_{var_base}"
                self.het_sites.append(new_var)
        if self.clip_3p_positions != [] and self.het_sites != []:
            clip_pos = max(self.clip_3p_positions)
            var_after_clip = [
                a for a in self.het_sites if int(a.split("_")[0]) > clip_pos
            ]
            if var_after_clip == []:
                var_pos = clip_pos + 30
                ref_base = self._refh.fetch(
                    self.nchr_old, var_pos - self.offset - 1, var_pos - self.offset
                ).upper()
                var_base = [a for a in ["A", "C", "G", "T"] if a != ref_base][0]
                new_var = f"{var_pos}_{ref_base}_{var_base}"
                self.het_sites.append(new_var)

        self.het_sites = sorted(self.het_sites)
        raw_read_haps = self.get_haplotypes_from_reads_step(
            exclude_reads,
            min_mapq,
            min_clip_len,
            check_clip,
            partial_deletion_reads,
            multi_allelic_sites,
        )
        return raw_read_haps

    def get_haplotypes_from_reads_step(
        self,
        exclude_reads=[],
        min_mapq=5,
        min_clip_len=50,
        check_clip=False,
        partial_deletion_reads=[],
        multi_allelic_sites={},
    ):
        """
        Go through reads and get bases at sites of interest
        Returns:
            read_haps (dict of str:list): collapse each read into just the positions
            of interest. 1 corresponds to ref, 2 corresponds to alt
        """
        het_sites = self.het_sites
        read_haps = {}
        nvar = len(het_sites)
        for dsnp_index, allele_site in enumerate(het_sites):
            snp_position_gene1, allele1, allele2, *at = allele_site.split("_")
            snp_position = int(snp_position_gene1)
            reads_with_flanking_indels = []
            for pileupcolumn in self._bamh.pileup(
                self.nchr,
                snp_position - 2,
                snp_position,
                truncate=True,
                min_base_quality=self.MEAN_BASE_QUAL,
            ):
                # require that the base on the read is not flanked by any indels
                if pileupcolumn.reference_pos == snp_position - 2:
                    for read in pileupcolumn.pileups:
                        if read.indel != 0 or read.is_del:
                            read_names = self.get_read_names(
                                read.alignment, partial_deletion_reads
                            )
                            for read_name in read_names:
                                reads_with_flanking_indels.append(read_name)
                if pileupcolumn.reference_pos == snp_position - 1:
                    for read in pileupcolumn.pileups:
                        read_names = self.get_read_names(
                            read.alignment, partial_deletion_reads
                        )
                        for read_name in read_names:
                            if (
                                not read.is_del
                                and not read.is_refskip
                                and not read.alignment.is_secondary
                                and read.alignment.mapping_quality >= min_mapq
                                and read_name not in exclude_reads
                                and read.indel == 0
                                and read_name not in reads_with_flanking_indels
                            ):
                                read_seq = read.alignment.query_sequence
                                start_pos = read.query_position
                                end_pos = start_pos + 1
                                if end_pos < len(read_seq):
                                    hap = read_seq[start_pos:end_pos]
                                    if read_name not in read_haps:
                                        read_haps.setdefault(read_name, ["x"] * nvar)
                                    if hap.upper() == allele1.upper() or (
                                        snp_position in multi_allelic_sites
                                        and hap.upper()
                                        == multi_allelic_sites[snp_position]
                                    ):
                                        read_haps[read_name][dsnp_index] = "1"
                                    elif hap.upper() == allele2.upper():
                                        read_haps[read_name][dsnp_index] = "2"

        # for softclips starting at a predefined position, mark sites as 0 instead of x
        if check_clip and (
            self.clip_3p_positions != [] or self.clip_5p_positions != []
        ):
            for dsnp_index, allele_site in enumerate(het_sites):
                snp_position_gene1, allele1, allele2, *at = allele_site.split("_")
                snp_position = int(snp_position_gene1)
                for clip_position in sorted(self.clip_3p_positions):
                    if snp_position > clip_position:
                        for read in self._bamh.fetch(
                            self.nchr, clip_position - 10, clip_position + 10
                        ):
                            read_name = self.get_read_name(read)
                            if read_name not in read_haps:
                                read_haps.setdefault(read_name, ["x"] * nvar)
                            if abs(read.reference_end - clip_position) < 20:
                                find_clip_3p = re.findall(
                                    self.clip_3p, read.cigarstring
                                )
                                if (
                                    find_clip_3p != []
                                    and int(find_clip_3p[0][:-1]) >= min_clip_len
                                ):
                                    read_haps[read_name][dsnp_index] = "0"
                for clip_position in sorted(self.clip_5p_positions, reverse=True):
                    if snp_position < clip_position:
                        for read in self._bamh.fetch(
                            self.nchr, clip_position - 10, clip_position + 10
                        ):
                            read_name = self.get_read_name(read)
                            if read_name not in read_haps:
                                read_haps.setdefault(read_name, ["x"] * nvar)
                            if abs(read.reference_start - clip_position) < 20:
                                find_clip_5p = re.findall(
                                    self.clip_5p, read.cigarstring
                                )
                                if (
                                    find_clip_5p != []
                                    and int(find_clip_5p[0][:-1]) >= min_clip_len
                                ):
                                    read_haps[read_name][dsnp_index] = "0"
        return read_haps

    def remove_var(self, raw_read_haps, kept_sites):
        """remove variants that are not present after checking each read-haplotype"""
        bases_per_site = {}
        sites_to_remove = []
        for i in range(len(self.het_sites)):
            for read, hap in raw_read_haps.items():
                base = hap[i]
                bases_per_site.setdefault(i, []).append(base)

        for pos in bases_per_site:
            bases = bases_per_site[pos]
            bases_x = bases.count("x")
            bases_ref = bases.count("1")
            bases_alt = bases.count("2")
            this_var = self.het_sites[pos]
            if bases_x == len(bases):
                sites_to_remove.append(this_var)
            elif bases_ref + bases_alt == len(bases) - bases_x and (
                bases_alt <= 3 or bases_ref <= 3
            ):
                if this_var not in kept_sites:
                    sites_to_remove.append(this_var)
        for var in sites_to_remove:
            if var in self.het_sites:
                self.het_sites.remove(var)

    def allow_del_bases(self, pos):
        """
        During variant calling, allow "bases in deletions" for positions in
        two long deletions
        """
        if (
            self.del2_reads_partial != set()
            and self.del2_3p_pos1 <= pos <= self.del2_5p_pos2
        ):
            return True
        if (
            self.del1_reads_partial != set()
            and self.del1_3p_pos1 <= pos <= self.del1_5p_pos2
        ):
            return True
        return False

    def process_indel(self, pos, ref_seq, var_seq):
        """Translate pysam indel seq into real sequence"""
        if "+" in var_seq:
            ins_base = var_seq.split(re.findall(r"\+\d+", var_seq)[0])[1]
            indel_size = len(ins_base)
            var_seq = ref_seq + ins_base
        else:
            del_len = int(re.findall(r"\-\d+", var_seq)[0][1:])
            indel_size = del_len
            var_seq = ref_seq
            offset_pos = pos - self.offset
            ref_seq = self._refh.fetch(
                self.nchr_old,
                offset_pos - 1,
                offset_pos + del_len,
            )
        return ref_seq, var_seq, indel_size

    def get_candidate_pos(
        self,
        regions_to_check=[],
        min_read_support=5,
        min_vaf=0.11,
        trusted_read_support=20,
        white_list={},
    ):
        """
        Get all polymorphic sites in the region, update self.candidate_pos
        """
        bamh = self._bamh
        pileups_raw = {}
        for pileupcolumn in bamh.pileup(
            self.nchr,
            self.left_boundary,
            self.right_boundary,
            truncate=True,
        ):
            pos = pileupcolumn.pos + 1
            pileups_raw.setdefault(
                pos,
                [a.upper() for a in pileupcolumn.get_query_sequences(add_indels=True)],
            )
        variants = {}
        variants_no_phasing = {}
        for pos in pileups_raw:
            all_bases = pileups_raw[pos]
            total_depth = len(all_bases)
            del_bases_count = all_bases.count("*")
            # get reference base
            offset_pos = pos - self.offset
            ref_seq_genome = self._refh.fetch(
                self.nchr_old, offset_pos - 1, offset_pos
            ).upper()

            if total_depth >= min_read_support and (
                del_bases_count < min_read_support or self.allow_del_bases(pos)
            ):
                all_bases = [a for a in all_bases if a != "*"]
                counter = Counter(all_bases)
                # include multi-allelic sites
                bases = counter.most_common(3)
                # homozygous
                if len(counter) == 1 or (
                    len(counter) >= 2
                    and bases[0][1]
                    > max(len(all_bases) - min_read_support, len(all_bases) * 0.85)
                ):
                    var_seq = bases[0][0]
                    ref_seq = ref_seq_genome
                    if var_seq != ref_seq:
                        # SNV and indels
                        if "-" not in var_seq and "+" not in var_seq:
                            # "homo" sites in large deletions should be put back into het sites
                            if (
                                self.allow_del_bases(pos)
                                and del_bases_count >= min_read_support
                                and pos not in self.homopolymer_sites
                            ):
                                variants.setdefault(pos, []).append((ref_seq, var_seq))
                            elif pos not in self.homopolymer_sites or (
                                pos in self.homopolymer_sites
                                and var_seq
                                not in self.homopolymer_sites[pos].split(",")
                            ):
                                self.homo_sites.append(f"{pos}_{ref_seq}_{var_seq}")
                        elif pos not in self.homopolymer_sites:
                            ref_seq, var_seq, indel_size = self.process_indel(
                                pos, ref_seq, var_seq
                            )
                            if indel_size < 25:
                                self.homo_sites.append(f"{pos}_{ref_seq}_{var_seq}")
                elif len(counter) >= 2:
                    found_ref = ref_seq_genome in [a[0] for a in bases]
                    if found_ref or pos in white_list:
                        for var_seq, var_count in bases:
                            ref_seq = ref_seq_genome
                            if var_seq != ref_seq and (
                                (
                                    var_count >= min_read_support
                                    and var_count / total_depth > min_vaf
                                )
                                or var_count >= trusted_read_support
                            ):
                                # SNV
                                if "-" not in var_seq and "+" not in var_seq:
                                    if pos not in self.homopolymer_sites:
                                        variants.setdefault(pos, []).append(
                                            (ref_seq, var_seq)
                                        )
                                    else:
                                        prohibited_bases = self.homopolymer_sites[
                                            pos
                                        ].split(",")
                                        if var_seq not in prohibited_bases:
                                            if "1" in prohibited_bases:
                                                variants.setdefault(pos, []).append(
                                                    (ref_seq, var_seq)
                                                )
                                            else:
                                                variants_no_phasing.setdefault(
                                                    pos, (ref_seq, var_seq)
                                                )
                                # indels
                                elif pos not in self.homopolymer_sites:
                                    ref_seq, var_seq, indel_size = self.process_indel(
                                        pos, ref_seq, var_seq
                                    )
                                    if indel_size < 25:
                                        variants_no_phasing.setdefault(
                                            pos, (ref_seq, var_seq)
                                        )

        # exclude variants caused by shifted softclips of the big deletions
        excluded_variants = []
        for region in regions_to_check:
            var_to_check = [a for a in variants if region[0] < a < region[1]]
            excluded_variants += var_to_check
        for pos in variants:
            # for now, filter out multi-allelic sites, except in white list
            if pos not in excluded_variants:
                if len(variants[pos]) == 1:
                    ref_seq, var_seq = variants[pos][0]
                    self.candidate_pos.add(f"{pos}_{ref_seq}_{var_seq}")
                elif pos in white_list:
                    for ref_seq, var_seq in variants[pos]:
                        if var_seq != white_list[pos]:
                            self.candidate_pos.add(f"{pos}_{ref_seq}_{var_seq}")
                            break

        excluded_variants = []
        for region in regions_to_check:
            var_to_check = [a for a in variants_no_phasing if region[0] < a < region[1]]
            excluded_variants += var_to_check
        for pos in variants_no_phasing:
            if pos not in excluded_variants:
                ref_seq, var_seq = variants_no_phasing[pos]
                self.het_no_phasing.append(f"{pos}_{ref_seq}_{var_seq}")

    def remove_noisy_sites(self):
        """remove variants in predefined noisy sites"""
        problematic_sites = []
        for site in self.het_sites:
            for region in self.noisy_region:
                if region[0] <= int(site.split("_")[0]) <= region[1]:
                    problematic_sites.append(site)
        for site in problematic_sites:
            self.het_sites.remove(site)

    @staticmethod
    def simplify_read_haps(read_haps):
        """Simplify read haplotypes for output"""
        haplotypes_to_reads = {}
        reads_to_haplotypes = {}
        for read in read_haps:
            hap = read_haps[read]
            haplotypes_to_reads.setdefault("".join(hap), []).append(read)
            reads_to_haplotypes.setdefault(read, "".join(hap))
        return haplotypes_to_reads, reads_to_haplotypes

    def check_variants_in_haplotypes(self, variant):
        """
        For variants not used in phasing, check which haplotypes they are in.
        """
        dreads = {}
        var_pos, ref, alt = variant.split("_")
        var_pos = int(var_pos)
        var_size = len(alt) - len(ref)
        if var_size < 0:
            indel_N = "N" * abs(var_size)
            indel_base_in_read = f"{ref[0]}{var_size}{indel_N}"
        elif var_size > 0:
            indel_base_in_read = f"{ref[0]}+{var_size}{alt[1:]}"
        ref_base = ref[0]
        alt_base = alt[0] if var_size == 0 else indel_base_in_read
        for pileupcolumn in self._bamh.pileup(
            self.nchr,
            var_pos - 1,
            var_pos,
            truncate=True,
        ):
            read_names = pileupcolumn.get_query_names()
            bases = [
                a.upper() for a in pileupcolumn.get_query_sequences(add_indels=True)
            ]
            for i, read in enumerate(read_names):
                if bases[i] == ref_base:
                    dreads.setdefault(read, "ref")
                elif bases[i] == alt_base:
                    dreads.setdefault(read, "alt")
                else:
                    dreads.setdefault(read, ".")
        return dreads

    @staticmethod
    def get_start_end(hap):
        """get range of positions that are not x"""
        haplen = len(hap)
        for nstart, base in enumerate(hap):
            if base != "x":
                break
        for nend in reversed(range(haplen)):
            if hap[nend] != "x":
                break
        return nstart, nend

    def get_hap_variant_ranges(self, hap):
        """get boundaries of (partial) haplotypes"""
        nstart, nend = self.get_start_end(hap)
        if nstart == 0:
            nstart_previous_pos = self.left_boundary
        else:
            nstart_previous = nstart - 1
            nstart_previous_pos = int(self.het_sites[nstart_previous].split("_")[0]) + 1
        if nend == len(hap) - 1:
            nend_next_pos = self.right_boundary
        else:
            nend_next = nend + 1
            nend_next_pos = int(self.het_sites[nend_next].split("_")[0]) - 1
        return nstart_previous_pos, nend_next_pos

    def output_variants_in_haplotypes(self, haps, reads, nonunique, known_del={}):
        """
        Summarize all variants in each haplotype.
        Output all variants and their genotypes.
        Haplotypes are different length, so a range (boundary) is reported
        """
        het_sites = self.het_sites
        haplotype_variants = {}
        haplotype_info = {}
        var_no_phasing = copy.deepcopy(self.het_no_phasing)
        for hap, hap_name in haps.items():
            haplotype_variants.setdefault(hap_name, [])
        # het sites not used in phasing
        if reads != {}:
            for var in var_no_phasing:
                genotypes = []
                var_reads = self.check_variants_in_haplotypes(var)
                haps_with_variant = []
                for hap, hap_name in haps.items():
                    hap_reads = reads[hap]
                    hap_reads_nonunique = [a for a in nonunique if hap in nonunique[a]]
                    genotype = self.get_genotype_in_hap(
                        var_reads, hap_reads, hap_reads_nonunique
                    )
                    genotypes.append(genotype)
                    if genotype == "1":
                        haps_with_variant.append(hap_name)
                if haps_with_variant == []:
                    self.het_no_phasing.remove(var)
                else:
                    for hap_name in haps_with_variant:
                        haplotype_variants[hap_name].append(var)
        # het sites and homo sites
        for hap, hap_name in haps.items():
            # find boundary for confident variant calling
            hap_bound_start, hap_bound_end = self.get_hap_variant_ranges(hap)
            is_truncated = False
            # het sites
            for i in range(len(hap)):
                if hap[i] == "2":
                    haplotype_variants[hap_name].append(het_sites[i])
                elif hap[i] in known_del:
                    del_name = known_del[hap[i]]
                    if del_name not in haplotype_variants[hap_name]:
                        haplotype_variants[hap_name].append(del_name)
            # homo sites
            filtered_homo_sites = self.homo_sites
            # check known deletions
            del1_name = None
            del2_name = None
            if known_del != {}:
                known_del_names = list(known_del.values())
                del1_name = known_del_names[0]
                if len(known_del_names) > 1:
                    del2_name = known_del_names[1]
            if del1_name in haplotype_variants[hap_name]:
                pos1 = self.del1_3p_pos1
                pos2 = self.del1_5p_pos2
                filtered_homo_sites = [
                    a
                    for a in filtered_homo_sites
                    if int(a.split("_")[0]) < pos1 or int(a.split("_")[0]) > pos2
                ]
            if del2_name in haplotype_variants[hap_name]:
                pos1 = self.del2_3p_pos1
                pos2 = self.del2_5p_pos2
                filtered_homo_sites = [
                    a
                    for a in filtered_homo_sites
                    if int(a.split("_")[0]) < pos1 or int(a.split("_")[0]) > pos2
                ]
            # haps with clips
            if hap.startswith("0") and self.clip_5p_positions != []:
                for first_pos_after_clip, base in enumerate(hap):
                    if base != "0":
                        break
                first_pos_after_clip_position = int(
                    het_sites[first_pos_after_clip].split("_")[0]
                )
                for clip_position in sorted(self.clip_5p_positions, reverse=True):
                    if clip_position < first_pos_after_clip_position:
                        break
                filtered_homo_sites = [
                    a
                    for a in filtered_homo_sites
                    if int(a.split("_")[0]) > clip_position
                ]
                hap_bound_start = max(hap_bound_start, clip_position)
                haplotype_variants[hap_name].append(f"{clip_position}_clip_5p")
                if clip_position > self.gene_start:
                    is_truncated = True
            if hap.endswith("0") and self.clip_3p_positions != []:
                for first_pos_before_clip in reversed(range(len(hap))):
                    if hap[first_pos_before_clip] != "0":
                        break
                first_pos_before_clip_position = int(
                    het_sites[first_pos_before_clip].split("_")[0]
                )
                for clip_position in sorted(self.clip_3p_positions):
                    if clip_position > first_pos_before_clip_position:
                        break
                filtered_homo_sites = [
                    a
                    for a in filtered_homo_sites
                    if int(a.split("_")[0]) < clip_position
                ]
                hap_bound_end = min(hap_bound_end, clip_position)
                haplotype_variants[hap_name].append(f"{clip_position}_clip_3p")
                if clip_position < self.gene_end:
                    is_truncated = True

            haplotype_variants[hap_name] += filtered_homo_sites

            hap_bound_start = max(hap_bound_start, self.left_boundary)
            hap_bound_end = min(hap_bound_end, self.right_boundary)
            boundary_gene2 = None
            if self.gene2_region is not None:
                bound1_in_other_gene = self.get_range_in_other_gene(hap_bound_start)
                bound2_in_other_gene = self.get_range_in_other_gene(hap_bound_end)
                if None not in [bound1_in_other_gene, bound2_in_other_gene]:
                    boundary_gene2 = [
                        min(bound1_in_other_gene, bound2_in_other_gene),
                        max(bound1_in_other_gene, bound2_in_other_gene),
                    ]
                else:
                    boundary_gene2 = [bound1_in_other_gene, bound2_in_other_gene]
            var_tmp = haplotype_variants[hap_name]
            var_tmp1 = [
                a
                for a in var_tmp
                if hap_bound_start <= int(a.split("_")[0]) <= hap_bound_end
            ]
            var_tmp1 = list(set(var_tmp1))
            var_tmp2 = sorted(var_tmp1, key=lambda x: int(x.split("_")[0]))
            haplotype_info.setdefault(
                hap_name,
                {
                    "variants": var_tmp2,
                    "boundary": [hap_bound_start, hap_bound_end],
                    "boundary_gene2": boundary_gene2,
                    "is_truncated": is_truncated,
                },
            )

        return haplotype_info

    def get_genotype_in_hap(self, var_reads, hap_reads, hap_reads_nonunique):
        """For a given variant, return its status in a haplotype"""
        hap_reads_contain_var = [var_reads[a] for a in hap_reads if a in var_reads]
        if len(hap_reads_contain_var) < 3:
            hap_reads_contain_var += [
                var_reads[a] for a in hap_reads_nonunique if a in var_reads
            ]
        if len(hap_reads_contain_var) >= 3:
            hap_reads_contain_var_counter = Counter(hap_reads_contain_var).most_common(
                2
            )
            if len(hap_reads_contain_var_counter) == 1 or hap_reads_contain_var_counter[
                1
            ][1] <= min(hap_reads_contain_var_counter[0][1] * 0.15, 2):
                if hap_reads_contain_var_counter[0][0] == "alt":
                    return "1"
                elif hap_reads_contain_var_counter[0][0] == "ref":
                    return "0"
        return "."

    @staticmethod
    def update_reads_for_deletions(
        raw_read_haps, het_sites, n1, n2, del_reads_partial, base, del_name
    ):
        """
        For reads carrying known big deletions, update read haplotype to
        reflect the deletion. This is needed for downstream phasing
        """
        pos1 = -1
        pos2 = -1
        for i, var in enumerate(het_sites):
            if int(var.split("_")[0]) > n1:
                pos1 = i
                break
        for i, var in enumerate(het_sites):
            if int(var.split("_")[0]) > n2:
                pos2 = i
                break
        if pos1 != -1 and pos2 != -1:
            if pos1 < pos2:
                for read in del_reads_partial:
                    if read in raw_read_haps:
                        hap = list(raw_read_haps[read])
                        for i in range(pos1, pos2):
                            hap[i] = base
                        raw_read_haps[read] = "".join(hap)
            elif pos1 == pos2:
                het_sites.insert(pos1, del_name)
                for read in raw_read_haps:
                    hap = list(raw_read_haps[read])
                    hap.insert(pos1, "x")
                    if read in del_reads_partial:
                        hap[pos1] = base
                    elif (
                        hap[pos1 - 1] == "0"
                        and pos1 - 1 >= 0
                        and hap[pos1 + 1] == "0"
                        and pos1 + 1 < len(hap)
                    ):
                        hap[pos1] = "0"
                    else:
                        flanking_left = hap[max(0, pos1 - 2) : pos1]
                        flanking_right = hap[
                            min(pos1 + 1, len(hap)) : min(pos1 + 3, len(hap))
                        ]
                        if "x" not in flanking_left and "x" not in flanking_right:
                            hap[pos1] = "1"
                    raw_read_haps[read] = "".join(hap)
        return raw_read_haps, het_sites

    def get_read_counts(self, uniquely_supporting_haps):
        """
        Get unique supporting read counts for each haplotype
        over a region where all haplotypes have defined bases
        """
        if uniquely_supporting_haps == {}:
            return {}
        nhap = len(uniquely_supporting_haps)
        nvar = len(self.het_sites)
        # depth per site for haplotype
        hap_bases = []
        for i in range(nhap):
            hap_bases.append([])
        j = 0
        for hap in uniquely_supporting_haps:
            reads = uniquely_supporting_haps[hap]
            for i in range(nvar):
                bases = [a[i] for a in reads]
                hap_bases[j].append(len(bases) - bases.count("x") - bases.count("0"))
            j += 1
        ranges = []
        for i in range(nvar):
            if min([hap_bases[a][i] for a in range(nhap)]) >= 5:
                for j in range(nvar):
                    if j > i:
                        if min([hap_bases[a][j] for a in range(nhap)]) < 5:
                            ranges.append([i, j])
                            break
                    if j == nvar - 1 and j > i:
                        ranges.append([i, nvar - 1])
        if ranges == []:
            return None
        longest_range = sorted(ranges, key=lambda x: x[1] - x[0], reverse=True)[0]
        mid = int((longest_range[1] + longest_range[0]) / 2)
        nstart = max(mid - 1, longest_range[0])
        nend = min(mid + 1, longest_range[1])
        if nend == nstart:
            return None

        read_count = {}
        for hap in uniquely_supporting_haps:
            reads = uniquely_supporting_haps[hap]
            lreads = [a for a in reads if a[nstart:nend] != "x" * (nend - nstart)]
            read_count.setdefault(hap, len(lreads))
        return read_count

    def phase_haps(self, raw_read_haps, min_support=4, debug=False):
        """
        Assemble and evaluate haplotypes
        """
        het_sites = self.het_sites
        haplotypes_to_reads, raw_read_haps = self.simplify_read_haps(raw_read_haps)

        ass_haps = []
        original_haps = []
        nvar = len(het_sites)
        hcn = 0
        if nvar == 0:
            return ([], [], 0, {}, {}, {}, None)
        elif nvar == 1:
            ass_haps = ["1", "2"]
            original_haps = ["1", "2"]
        else:
            pivot_index, _ = self.get_pivot_site_index()
            hap_graph = VariantGraph(
                raw_read_haps, pivot_index, figure_id=self.sample_id
            )
            ass_haps, original_haps, hcn = hap_graph.run(debug=debug, make_plot=debug)

        if ass_haps == []:
            return (ass_haps, original_haps, hcn, {}, {}, raw_read_haps, None)

        (
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            read_counts,
        ) = self.get_read_support(raw_read_haps, haplotypes_to_reads, ass_haps)

        # remove spurious ones
        ass_haps = self.adjust_spurious_haplotypes(uniquely_supporting_reads)
        (
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            read_counts,
        ) = self.get_read_support(raw_read_haps, haplotypes_to_reads, ass_haps)

        # remove low-support ones
        read_counts = [
            len(uniquely_supporting_reads[a]) for a in uniquely_supporting_reads
        ]
        read_counts = sorted(read_counts)
        # more stringent if one copy is low but others are high
        if (
            len(read_counts) > 2
            and read_counts[0] <= 4
            and read_counts[1] >= 12
            and "x" not in "".join(ass_haps)
        ):
            min_support = 5

        ass_haps = [
            a
            for a in uniquely_supporting_reads
            if len(uniquely_supporting_reads[a]) >= min_support
        ]
        (
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            read_counts,
        ) = self.get_read_support(raw_read_haps, haplotypes_to_reads, ass_haps)

        return (
            ass_haps,
            original_haps,
            hcn,
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            read_counts,
        )

    def get_read_support(self, raw_read_haps, haplotypes_to_reads, ass_haps):
        """Find uniquely and nonuniquely supporting reads for given haplotypes"""
        read_support = VariantGraph.match_reads_and_haplotypes(raw_read_haps, ass_haps)
        uniquely_supporting_haps = read_support.unique
        read_counts = self.get_read_counts(uniquely_supporting_haps)

        uniquely_supporting_reads = {}
        for hap in ass_haps:
            uniquely_supporting_reads.setdefault(hap, [])
        for hap in uniquely_supporting_haps:
            for read_hap in uniquely_supporting_haps[hap]:
                uniquely_supporting_reads[hap] += haplotypes_to_reads[read_hap]
        for hap in uniquely_supporting_haps:
            uniquely_supporting_reads[hap] = list(set(uniquely_supporting_reads[hap]))

        nonuniquely_supporting_reads = {}
        for read in read_support.by_read:
            num_matches = len(read_support.by_read[read])
            if num_matches > 1:
                nonuniquely_supporting_reads.setdefault(
                    read, read_support.by_read[read]
                )
        return (
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            read_counts,
        )

    def compare_depth(self, haplotypes, loose=False, stringent=False):
        """
        For each haplotype, identify the variants where it's different
        from other haplotypes. Check depth at those variant sites and
        see if the depth suggests twice coverage.
        """
        if haplotypes is None or len(haplotypes) == 1:
            return []
        two_cp_haps = []
        bamh = self._bamh
        boundaries = [haplotypes[a]["boundary"] for a in haplotypes]
        nstart = max([a[0] for a in boundaries])
        nend = min(a[1] for a in boundaries)
        variants = set()
        for hap in haplotypes:
            vars = haplotypes[hap]["variants"]
            for var in vars:
                if len(var.split("_")) == 3:
                    pos, ref, alt = var.split("_")
                    pos = int(pos)
                    if nstart < pos < nend and var in self.het_sites:
                        variants.add(var)

        for hap in haplotypes:
            sites = {}
            other_haps = [a for a in haplotypes.keys() if a != hap]
            other_cn = len(other_haps)
            this_hap_var = haplotypes[hap]["variants"]
            other_haps_var = []
            for a in other_haps:
                other_haps_var += haplotypes[a]["variants"]
            for var in variants:
                pos, ref, alt = var.split("_")
                pos = int(pos)
                if var in this_hap_var and var not in other_haps_var:
                    sites.setdefault(pos, alt)
                elif var not in this_hap_var and other_haps_var.count(var) == other_cn:
                    sites.setdefault(pos, ref)

            counts = []
            for pos in sites:
                hap_base = sites[pos]
                for pileupcolumn in bamh.pileup(
                    self.nchr,
                    pos - 1,
                    pos,
                    truncate=True,
                    min_base_quality=self.MEAN_BASE_QUAL,
                ):
                    bases = [a.upper() for a in pileupcolumn.get_query_sequences()]
                    base_num = bases.count(hap_base)
                    counts.append([base_num, len(bases) - base_num])

            probs = []
            nsites = len(sites)
            for n1, n2 in counts:
                probs.append(self.depth_prob(n1, n2 / other_cn))
            probs_fil = [a for a in probs if a[0] < 0.25]
            if stringent is True:
                if len(probs_fil) >= nsites * 0.8 and nsites >= 5:
                    two_cp_haps.append(hap)
            elif len(probs_fil) >= nsites * 0.6 and nsites >= 5:
                two_cp_haps.append(hap)
            elif loose is True:
                if len(probs_fil) >= nsites * 0.5 and nsites >= 5:
                    two_cp_haps.append(hap)

        # there can only be one such haplotype
        if len(two_cp_haps) > 1:
            return []
        return two_cp_haps

    def adjust_spurious_haplotypes(self, uniquely_supporting_reads, flanking_bp=10):
        """Identify spurious haplotypes caused by locally misaligned reads"""
        passing_haplotypes = list(uniquely_supporting_reads.keys())
        suspicious_hap_pair = []
        lhap = uniquely_supporting_reads.keys()
        for hap1 in lhap:
            for hap2 in lhap:
                if hap1 != hap2:
                    nmatch = 0
                    nmismatch = 0
                    mismatch_sites = []
                    for i, base1 in enumerate(hap1):
                        base2 = hap2[i]
                        if "x" not in [base1, base2]:
                            if base1 == base2:
                                nmatch += 1
                            elif base1 in ["1", "2"] and base2 in ["1", "2"]:
                                nmismatch += 1
                                mismatch_sites.append(self.het_sites[i])
                    if nmatch >= 5 and nmismatch == 1 and len(mismatch_sites) == 1:
                        mismatch_pos = int(mismatch_sites[0].split("_")[0])
                        hap1_reads = uniquely_supporting_reads[hap1]
                        hap2_reads = uniquely_supporting_reads[hap2]
                        hap_pair = None
                        if len(hap1_reads) <= 5 and len(hap2_reads) >= 6:
                            hap_pair = [hap2, hap1]
                        elif len(hap2_reads) <= 5 and len(hap1_reads) >= 6:
                            hap_pair = [hap1, hap2]
                        if (
                            hap_pair is not None
                            and [
                                hap_pair,
                                mismatch_pos,
                            ]
                            not in suspicious_hap_pair
                        ):
                            suspicious_hap_pair.append([hap_pair, mismatch_pos])

        for hap_pair, mismatch_pos in suspicious_hap_pair:
            hap1, hap2 = hap_pair
            hap1_reads = uniquely_supporting_reads[hap1]
            hap2_reads = uniquely_supporting_reads[hap2]
            hap1_reads_at_pos = []
            hap2_reads_at_pos = []
            for pileupcolumn in self._bamh.pileup(
                self.nchr,
                mismatch_pos - 1,
                mismatch_pos,
                truncate=True,
                min_base_quality=self.MEAN_BASE_QUAL,
            ):
                for read in pileupcolumn.pileups:
                    if not read.is_del and not read.is_refskip:
                        read_name = self.get_read_name(read.alignment)
                        read_seq = read.alignment.query_sequence
                        read_pos = read.query_position
                        if (
                            read_name in hap1_reads + hap2_reads
                            and read_pos >= flanking_bp
                            and read_pos + flanking_bp < len(read_seq)
                        ):
                            start_pos = read_pos - flanking_bp
                            end_pos = read_pos + flanking_bp
                            if read_name in hap1_reads:
                                hap1_reads_at_pos.append(read_seq[start_pos:end_pos])
                            if read_name in hap2_reads:
                                hap2_reads_at_pos.append(read_seq[start_pos:end_pos])

            if set(hap1_reads_at_pos).intersection(set(hap2_reads_at_pos)) != set():
                if hap2 in passing_haplotypes:
                    passing_haplotypes.remove(hap2)
        return passing_haplotypes

    @staticmethod
    def check_linking_read(aln1, aln2, reverse=False):
        """Determine the direction of links between two alignments"""
        if reverse is False:
            if aln1[-1] != "x" and aln2[0] != "x":
                return "1-2"
            if aln2[-1] != "x" and aln1[0] != "x":
                return "2-1"
        else:
            if aln1[-1] != "x" and aln2[-1] != "x":
                return "0-0"
            if aln1[0] != "x" and aln2[0] != "x":
                return "0-0"
        return None

    def phase_alleles(
        self,
        uniq_reads,
        nonuniquely_supporting_reads,
        raw_read_haps,
        ass_haps,
        reverse=False,
        min_read=2,
    ):
        """
        Phase haplotypes into alleles using read evidence
        """
        new_reads = {}
        # unique
        for hap in uniq_reads:
            for read in uniq_reads[hap]:
                short_name = read.split("_sup")[0]
                new_reads.setdefault(short_name, []).append({read: [hap]})
        # nonunique
        for read, supported_haps in nonuniquely_supporting_reads.items():
            short_name = read.split("_sup")[0]
            new_reads.setdefault(short_name, []).append({read: supported_haps})

        (
            nondirected_links,
            directed_links,
            directed_links_loose,
        ) = self.get_directed_links(new_reads, raw_read_haps, ass_haps, reverse)

        read_links = {}
        for hap_link in nondirected_links:
            if len(nondirected_links[hap_link]) >= min_read:
                hap1, hap2 = hap_link.split("-")
                read_links.setdefault(hap1, []).append(hap2)
                read_links.setdefault(hap2, []).append(hap1)
        read_links = dict(
            sorted(read_links.items(), key=lambda item: len(item[1]), reverse=True)
        )
        alleles = Phaser.get_alleles_from_links(read_links, ass_haps.values())
        return (
            alleles,
            read_links,
            {a: len(b) for a, b in directed_links.items()},
            {a: Counter(b) for a, b in directed_links_loose.items()},
        )

    def get_directed_links(self, new_reads, raw_read_haps, ass_haps, reverse):
        """Get links between haplotypes from reads"""
        nondirected_links = {}
        directed_links = {}
        directed_links_loose = {}
        for read, hap_info in new_reads.items():
            nsegment = len(hap_info)
            if nsegment >= 2:
                links_found = set()
                for i in range(nsegment):
                    for j in range(i + 1, nsegment):
                        hap_info1 = hap_info[i]
                        hap_info2 = hap_info[j]
                        read1, haps1 = list(hap_info1.items())[0]
                        read2, haps2 = list(hap_info2.items())[0]
                        aln1 = raw_read_haps[read1]
                        aln2 = raw_read_haps[read2]
                        check_link = self.check_linking_read(aln1, aln2, reverse)
                        if len(haps1) == 1 and len(haps2) == 1:
                            hap1 = haps1[0]
                            hap2 = haps2[0]
                            if hap1 != hap2:
                                hap1_renamed = ass_haps[hap1]
                                hap2_renamed = ass_haps[hap2]
                                link_to_add = f"{hap1_renamed}-{hap2_renamed}"
                                if link_to_add not in links_found:
                                    links_found.add(link_to_add)
                                    nondirected_links.setdefault(
                                        link_to_add, []
                                    ).append(1)
                                if check_link is not None and check_link in [
                                    "1-2",
                                    "0-0",
                                ]:
                                    directed_links.setdefault(
                                        f"{hap1_renamed}-{hap2_renamed}", []
                                    ).append(1)
                                    directed_links_loose.setdefault(
                                        hap1_renamed, []
                                    ).append(hap2_renamed)
                                elif check_link is not None and check_link in [
                                    "2-1",
                                    "0-0",
                                ]:
                                    directed_links.setdefault(
                                        f"{hap2_renamed}-{hap1_renamed}", []
                                    ).append(1)
                                    directed_links_loose.setdefault(
                                        hap2_renamed, []
                                    ).append(hap1_renamed)
                        elif len(haps1) == 1 or len(haps2) == 1:
                            for hap1 in haps1:
                                for hap2 in haps2:
                                    if hap1 != hap2:
                                        hap1_renamed = ass_haps[hap1]
                                        hap2_renamed = ass_haps[hap2]
                                        if check_link is not None and check_link in [
                                            "1-2",
                                            "0-0",
                                        ]:
                                            directed_links_loose.setdefault(
                                                hap1_renamed, []
                                            ).append(hap2_renamed)
                                        elif check_link is not None and check_link in [
                                            "2-1",
                                            "0-0",
                                        ]:
                                            directed_links_loose.setdefault(
                                                hap2_renamed, []
                                            ).append(hap1_renamed)
        return nondirected_links, directed_links, directed_links_loose

    @staticmethod
    def get_alleles_from_links(read_links, total_haps):
        """phase alleles from read_links between reads"""
        alleles = []
        if read_links != {}:
            alleles = [[list(read_links.keys())[0]] + list(read_links.values())[0]]
            for hap1 in read_links:
                for hap2 in read_links[hap1]:
                    hap1_in = sum([hap1 in a for a in alleles])
                    hap2_in = sum([hap2 in a for a in alleles])
                    if hap1_in == 0 and hap2_in == 0:
                        alleles.append([hap1, hap2])
                    elif hap1_in == 0:
                        for a in alleles:
                            if hap2 in a:
                                a_index = alleles.index(a)
                                if hap1 not in alleles[a_index]:
                                    alleles[a_index].append(hap1)
                                if hap2 not in alleles[a_index]:
                                    alleles[a_index].append(hap2)
                    else:
                        for a in alleles:
                            if hap1 in a:
                                a_index = alleles.index(a)
                                if hap2 not in alleles[a_index]:
                                    alleles[a_index].append(hap2)
                                if hap1 not in alleles[a_index]:
                                    alleles[a_index].append(hap1)
        # merge alleles
        while True:
            to_merge = []
            for hap in total_haps:
                hap_found_in_alleles = [hap in a for a in alleles]
                if hap_found_in_alleles.count(True) > 1:
                    to_merge.append(hap)
                    break
            if to_merge == []:
                break
            new_alleles = []
            hap = to_merge[0]
            merged = []
            for each_allele in alleles:
                if hap not in each_allele:
                    new_alleles.append(each_allele)
                else:
                    merged += each_allele
            merged = list(set(merged))
            new_alleles.append(merged)
            alleles = new_alleles
        return alleles

    def add_homo_sites(self, min_no_var_region_size=10000, max_homo_var_to_add=10):
        """add some homozygous sites to het sites in long homozygous regions"""
        if self.het_sites == []:
            return []
        het_pos = [int(a.split("_")[0]) for a in self.het_sites]
        min_pos = min(het_pos)
        max_pos = max(het_pos)
        homo_sites_to_add = []
        if min_pos - self.left_boundary > min_no_var_region_size:
            for var in self.homo_sites:
                pos, ref, alt = var.split("_")
                if int(pos) < min_pos and len(ref) == 1 and len(alt) == 1:
                    homo_sites_to_add.append(var)
        if self.right_boundary - max_pos > min_no_var_region_size:
            for var in self.homo_sites:
                pos, ref, alt = var.split("_")
                if int(pos) > max_pos and len(ref) == 1 and len(alt) == 1:
                    homo_sites_to_add.append(var)
        if homo_sites_to_add != []:
            homo_sites_to_add = sorted(homo_sites_to_add)
            homo_sites_to_add_size = len(homo_sites_to_add)
            homo_sites_to_add = homo_sites_to_add[
                :: int(np.ceil(homo_sites_to_add_size / max_homo_var_to_add))
            ]
            self.het_sites += homo_sites_to_add
            self.het_sites = sorted(self.het_sites)
            self.remove_noisy_sites()
        return homo_sites_to_add

    def call(self):
        """Main function to phase haplotypes and call copy numbers"""
        if self.check_coverage_before_analysis() is False:
            return self.GeneCall()
        self.get_homopolymer()
        # self.get_homopolymer_old()
        self.find_big_deletion()

        if self.deletion1_size is not None:
            self.del1_reads, self.del1_reads_partial = self.get_long_del_reads(
                self.del1_3p_pos1,
                self.del1_3p_pos2,
                self.del1_5p_pos1,
                self.del1_5p_pos2,
                self.deletion1_size,
            )
        if self.deletion2_size is not None:
            self.del2_reads, self.del2_reads_partial = self.get_long_del_reads(
                self.del2_3p_pos1,
                self.del2_3p_pos2,
                self.del2_5p_pos1,
                self.del2_5p_pos2,
                self.deletion2_size,
            )

        # scan for polymorphic sites
        regions_to_check = []
        if self.del2_reads_partial != set():
            regions_to_check += [
                [self.del2_3p_pos1, self.del2_3p_pos2],
                [self.del2_5p_pos1, self.del2_5p_pos2],
            ]
        if self.del1_reads_partial != set():
            regions_to_check += [
                [self.del1_3p_pos1, self.del1_3p_pos2],
                [self.del1_5p_pos1, self.del1_5p_pos2],
            ]
        self.get_candidate_pos(regions_to_check=regions_to_check)

        self.het_sites = sorted(list(self.candidate_pos))
        self.remove_noisy_sites()
        homo_sites_to_add = self.add_homo_sites()

        raw_read_haps = self.get_haplotypes_from_reads(
            check_clip=True, kept_sites=homo_sites_to_add, add_sites=self.add_sites
        )

        het_sites = self.het_sites
        known_del = {}
        if self.del1_reads_partial != set():
            raw_read_haps, het_sites = self.update_reads_for_deletions(
                raw_read_haps,
                het_sites,
                self.del1_3p_pos1,
                self.del1_5p_pos2,
                self.del1_reads_partial,
                "3",
                self.deletion1_name,
            )
            known_del.setdefault("3", self.deletion1_name)
        if self.del2_reads_partial != set():
            raw_read_haps, het_sites = self.update_reads_for_deletions(
                raw_read_haps,
                het_sites,
                self.del2_3p_pos1,
                self.del2_5p_pos2,
                self.del2_reads_partial,
                "4",
                self.deletion2_name,
            )
            known_del.setdefault("4", self.deletion2_name)
        self.het_sites = het_sites

        (
            ass_haps,
            original_haps,
            hcn,
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            read_counts,
        ) = self.phase_haps(raw_read_haps)

        tmp = {}
        mod_gene_name = ",".join(self.gene.split("-"))
        for i, hap in enumerate(ass_haps):
            tmp.setdefault(hap, f"{mod_gene_name}_hap{i+1}")
        ass_haps = tmp

        haplotypes = None
        if ass_haps != {}:
            haplotypes = self.output_variants_in_haplotypes(
                ass_haps,
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
                known_del=known_del,
            )

        two_cp_haps = []
        if len(ass_haps) == 3:
            two_cp_haps = self.compare_depth(haplotypes, stringent=True)
            if two_cp_haps == [] and read_counts is not None:
                # check if one haplotype has more reads than others
                haps = list(read_counts.keys())
                counts = list(read_counts.values())
                max_count = max(counts)
                cp2_hap = haps[counts.index(max_count)]
                others_max = sorted(counts, reverse=True)[1]
                probs = self.depth_prob(max_count, others_max)
                # print(counts, probs)
                if probs[0] < 0.05 and others_max >= 10:
                    two_cp_haps.append(ass_haps[cp2_hap])

        total_cn = len(ass_haps) + len(two_cp_haps)

        # fully homozygous
        if self.het_sites == [] or total_cn == 1:
            total_cn = 2

        # two pairs of identical copies
        if two_cp_haps == [] and total_cn == 2 and self.expect_cn2 is False:
            if self.mdepth is not None:
                prob = self.depth_prob(int(self.region_avg_depth.median), self.mdepth)
                if prob[0] < 0.75:
                    total_cn = 4

        # phase
        alleles = []
        hap_links = {}
        if self.to_phase is True:
            (
                alleles,
                hap_links,
                _,
                _,
            ) = self.phase_alleles(
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
                raw_read_haps,
                ass_haps,
                reverse=self.is_reverse,
            )
        self.close_handle()

        return self.GeneCall(
            total_cn,
            None,
            ass_haps,
            two_cp_haps,
            alleles,
            hap_links,
            hcn,
            original_haps,
            self.het_sites,
            uniquely_supporting_reads,
            self.het_no_phasing,
            self.homo_sites,
            haplotypes,
            nonuniquely_supporting_reads,
            raw_read_haps,
            self.mdepth,
            self.region_avg_depth._asdict(),
            self.sample_sex,
        )

    def close_handle(self):
        self._bamh.close()
        self._refh.close()



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>


from collections import namedtuple
import copy
from ..phaser import Phaser


class RccxPhaser(Phaser):
    fields = (
        "total_cn",
        "final_haplotypes",
        "two_copy_haplotypes",
        "starting_hap",
        "ending_hap",
        "deletion_hap",
        "phasing_success",
        "alleles_final",
        "annotated_alleles",
        "hap_variants",
        "hap_links",
        "highest_total_cn",
        "assembled_haplotypes",
        "sites_for_phasing",
        "unique_supporting_reads",
        "het_sites_not_used_in_phasing",
        "homozygous_sites",
        "haplotype_details",
        "nonunique_supporting_reads",
        "read_details",
        "genome_depth",
        "region_depth",
    )
    GeneCall = namedtuple(
        "GeneCall",
        fields,
        defaults=(None,) * len(fields),
    )

    def __init__(
        self, sample_id, outdir, genome_depth=None, genome_bam=None, sample_sex=None
    ):
        Phaser.__init__(self, sample_id, outdir, genome_depth, genome_bam, sample_sex)
        self.has_gene1 = False
        self.has_gene2 = False
        self.gene1_reads = set()
        self.gene2_reads = set()
        self.del1_reads = set()
        self.del1_reads_partial = set()
        self.del2_reads = set()
        self.del2_reads_partial = set()

    def set_parameter(self, config):
        super().set_parameter(config)
        self.variant_def = config["data"]["snp_file"]
        self.known_variants = {}
        with open(self.variant_def) as f:
            for line in f:
                split_line = line.split()
                self.known_variants.setdefault(
                    "_".join([split_line[1], split_line[2], split_line[3]]),
                    split_line[-1],
                )
        self.white_list = config["white_list"]
        self.deletion1_size = config["deletion1_size"]
        self.deletion2_size = config["deletion2_size"]
        self.deletion1_name = config["deletion1_name"]
        self.deletion2_name = config["deletion2_name"]
        self.del2_3p_pos1 = config["del2_3p_pos1"]
        self.del2_3p_pos2 = config["del2_3p_pos2"]
        self.del2_5p_pos1 = config["del2_5p_pos1"]
        self.del2_5p_pos2 = config["del2_5p_pos2"]
        self.del1_3p_pos1 = config["del1_3p_pos1"]
        self.del1_3p_pos2 = config["del1_3p_pos2"]
        self.del1_5p_pos1 = config["del1_5p_pos1"]
        self.del1_5p_pos2 = config["del1_5p_pos2"]

    def allow_del_bases(self, pos):
        """
        During variant calling, allow "bases in deletions" for positions in
        two long deletions
        """
        if (
            self.del2_reads_partial != set()
            and self.del2_3p_pos1 <= pos <= self.del2_5p_pos2
        ):
            return True
        if (
            self.del1_reads_partial != set()
            and self.del1_3p_pos1 <= pos <= self.del1_5p_pos2
        ):
            return True
        return False

    @staticmethod
    def annotate_var(allele_var):
        """annotate an allele with variants"""
        annotated_allele = None
        if len(allele_var) == 2:
            if [] in allele_var:
                annotated_allele = "WT"
            else:
                tmp = sorted(allele_var, key=lambda x: len(x))
                annotated_allele = ",".join(tmp[0])
        elif len(allele_var) == 1:
            if allele_var == [[]]:
                annotated_allele = "pseudogene_deletion"
            else:
                annotated_allele = "deletion_" + ",".join(allele_var[0])
        elif len(allele_var) == 3:
            tmp = sorted(allele_var, key=lambda x: len(x))
            if tmp[0] == []:
                if tmp[1] == []:
                    annotated_allele = "gene_duplication"
                elif abs(len(tmp[1]) - len(tmp[2])) <= 1 and len(tmp[2]) >= 6:
                    annotated_allele = "pseudogene_duplication"
                else:
                    annotated_allele = "duplication_WT_plus_" + ",".join(tmp[1])
            else:
                if abs(len(tmp[1]) - len(tmp[2])) <= 1 and len(tmp[2]) >= 6:
                    annotated_allele = ",".join(tmp[0]) + "_pseudogene_duplication"
                else:
                    annotated_allele = (
                        "duplication_" + ",".join(tmp[0]) + "_plus_" + ",".join(tmp[1])
                    )

        return annotated_allele

    def annotate_alleles(
        self,
        successful_phasing,
        new_alleles,
        hap_variants,
        ending_copies,
        ass_haps,
        two_cp_haplotypes,
    ):
        """annotate the allele type"""
        annotated_alleles = []
        if successful_phasing:
            allele1 = new_alleles[0]
            allele2 = new_alleles[1]
            allele1_var = [hap_variants[a] for a in allele1]
            allele2_var = [hap_variants[a] for a in allele2]
            for allele_var in [allele1_var, allele2_var]:
                annotated_allele = self.annotate_var(allele_var)
                annotated_alleles.append(annotated_allele)
        else:
            if (
                len(ending_copies) == 2
                and len(ass_haps) == 4
                and two_cp_haplotypes == []
            ):
                for allele_var in [hap_variants[a] for a in ending_copies]:
                    annotated_allele = None
                    if allele_var == []:
                        annotated_allele = "WT"
                    else:
                        annotated_allele = ",".join(allele_var)
                    annotated_alleles.append(annotated_allele)
        return annotated_alleles

    def update_alleles(
        self,
        new_alleles,
        haplotypes,
        final_haps,
        single_copies,
        starting_copies,
        ending_copies,
        hcn,
    ):
        """Update phased alleles"""
        two_cp_haplotypes = []
        successful_phasing = False
        # the deletion haplotype will be reported as an allele
        if len(single_copies) == 1 and len(final_haps) < 5:
            if (
                len(new_alleles) == 1
                and len(new_alleles[0]) == len(final_haps) - 1
                and single_copies not in new_alleles
            ):
                new_alleles.append(single_copies)
            elif (
                len(new_alleles) == 1
                and len(new_alleles[0]) < len(final_haps) - 1
                and single_copies not in new_alleles
                and len(starting_copies) == 1
                and len(ending_copies) == 1
            ):
                new_alleles = []
            if new_alleles == []:
                new_alleles.append(single_copies)
                remaining_hap = [
                    a for a in final_haps.values() if a not in single_copies
                ]
                if len(remaining_hap) == len(final_haps) - 1:
                    new_alleles.append(remaining_hap)
        # deletions on each allele
        elif (
            len(single_copies) == 2
            and len(new_alleles) == 0
            and len(final_haps) == 2
            and len(starting_copies) == 0
            and len(ending_copies) == 0
        ):
            new_alleles = [[single_copies[0]], [single_copies[1]]]
            successful_phasing = True
        elif single_copies == []:
            # homozygous, each haplotype has cn 2
            if (
                len(final_haps) == 2
                and len(ending_copies) == 1
                and len(starting_copies) == 1
                and len(new_alleles) == 1
            ):
                two_cp_haplotypes = list(final_haps.values())
                new_alleles.append(new_alleles[0])
                successful_phasing = True
            # depth-based adjustment when found 3 haplotypes or <2 ending haplotypes
            if haplotypes is not None:
                two_cp_hap_candidate = self.compare_depth(haplotypes, loose=True)
                if len(ending_copies) == 1 and len(starting_copies) == 2:
                    if two_cp_hap_candidate == ending_copies:
                        two_cp_haplotypes = two_cp_hap_candidate
                        if len(final_haps) == 3:
                            new_alleles = [
                                [starting_copies[0], ending_copies[0]],
                                [starting_copies[1], ending_copies[0]],
                            ]
                            successful_phasing = True
                elif len(starting_copies) == 1 and len(ending_copies) == 2:
                    if two_cp_hap_candidate == starting_copies:
                        two_cp_haplotypes = two_cp_hap_candidate
                        if len(final_haps) == 3:
                            new_alleles = [
                                [starting_copies[0], ending_copies[0]],
                                [starting_copies[0], ending_copies[1]],
                            ]
                            successful_phasing = True

            # add missing links when there is no two-cp haplotypes
            if two_cp_haplotypes == [] and len(ending_copies) <= 2:
                # add the missing link in cn=4
                if (
                    len(final_haps) in [3, 4]
                    and len(new_alleles) == 1
                    and len(new_alleles[0]) == 2
                    and hcn == len(final_haps)
                ):
                    remaining_hap = [
                        a for a in final_haps.values() if a not in new_alleles[0]
                    ]
                    if len(remaining_hap) == len(final_haps) - 2:
                        new_alleles.append(remaining_hap)
                # add the missing link in cn=5
                if len(final_haps) == 5:
                    if (
                        len(new_alleles) == 1
                        and len(new_alleles[0]) == 2
                        and (
                            (
                                new_alleles[0][0] in starting_copies
                                and new_alleles[0][1] in ending_copies
                            )
                            or (
                                new_alleles[0][1] in starting_copies
                                and new_alleles[0][0] in ending_copies
                            )
                        )
                    ):
                        remaining_hap = [
                            a for a in final_haps.values() if a not in new_alleles[0]
                        ]
                        if len(remaining_hap) == 3:
                            new_alleles.append(remaining_hap)
                    if len(new_alleles) == 2:
                        allele1 = (
                            new_alleles[0][0] in starting_copies
                            and new_alleles[0][1] in ending_copies
                        ) or (
                            new_alleles[0][1] in starting_copies
                            and new_alleles[0][0] in ending_copies
                        )
                        allele2 = (
                            new_alleles[1][0] in starting_copies
                            and new_alleles[1][1] in ending_copies
                        ) or (
                            new_alleles[1][1] in starting_copies
                            and new_alleles[1][0] in ending_copies
                        )
                        if allele1 is True and allele2 is False:
                            remaining_hap = [
                                a
                                for a in final_haps.values()
                                if a not in new_alleles[0]
                            ]
                            if len(remaining_hap) == 3:
                                new_alleles = [new_alleles[0], remaining_hap]
                        elif allele1 is False and allele2 is True:
                            remaining_hap = [
                                a
                                for a in final_haps.values()
                                if a not in new_alleles[1]
                            ]
                            if len(remaining_hap) == 3:
                                new_alleles = [new_alleles[1], remaining_hap]

        # check wrong phasing
        wrong_allele = False
        for allele in new_alleles:
            hp_set = set(allele)
            # if both copies are starting or ending
            if len(allele) == 2:
                hp1, hp2 = allele
                if (hp1 in starting_copies and hp2 in starting_copies) or (
                    hp1 in ending_copies and hp2 in ending_copies
                ):
                    wrong_allele = True
                    break
            for hp in hp_set:
                if (
                    hp in ending_copies
                    and allele.count(hp) > 1
                    and hp not in two_cp_haplotypes
                ):
                    wrong_allele = True
                    break
        if wrong_allele:
            new_alleles = []

        if len(new_alleles) == 2:
            if sorted(new_alleles[0] + new_alleles[1]) == sorted(
                list(final_haps.values())
            ):
                successful_phasing = True

        return successful_phasing, new_alleles, two_cp_haplotypes

    def call(self):
        if self.check_coverage_before_analysis() is False:
            return self.GeneCall()
        self.get_homopolymer()
        self.del2_reads, self.del2_reads_partial = self.get_long_del_reads(
            self.del2_3p_pos1,
            self.del2_3p_pos2,
            self.del2_5p_pos1,
            self.del2_5p_pos2,
            self.deletion2_size,
        )
        self.del1_reads, self.del1_reads_partial = self.get_long_del_reads(
            self.del1_3p_pos1,
            self.del1_3p_pos2,
            self.del1_5p_pos1,
            self.del1_5p_pos2,
            self.deletion1_size,
        )

        # scan for polymorphic sites
        regions_to_check = []
        if self.del2_reads_partial != set():
            regions_to_check += [
                [self.del2_3p_pos1, self.del2_3p_pos2],
                [self.del2_5p_pos1, self.del2_5p_pos2],
            ]
        if self.del1_reads_partial != set():
            regions_to_check += [
                [self.del1_3p_pos1, self.del1_3p_pos2],
                [self.del1_5p_pos1, self.del1_5p_pos2],
            ]
        self.get_candidate_pos(
            regions_to_check=regions_to_check,
            white_list=self.white_list,
        )

        # add last snp outside of repeat
        var_found = False
        for var in self.candidate_pos:
            pos = int(var.split("_")[0])
            if pos > self.clip_3p_positions[0]:
                var_found = True
                break
        if var_found is False and self.candidate_pos != set():
            self.candidate_pos.add(self.add_sites[1])
        # add last snp outside of repeat, 5prime
        var_found = False
        for var in self.candidate_pos:
            pos = int(var.split("_")[0])
            if pos < self.clip_5p_positions[0]:
                var_found = True
                break
        if var_found is False and self.candidate_pos != set():
            self.candidate_pos.add(self.add_sites[0])

        self.het_sites = sorted(list(self.candidate_pos))
        self.remove_noisy_sites()

        raw_read_haps = self.get_haplotypes_from_reads(
            check_clip=True,
            partial_deletion_reads=self.del1_reads_partial,
            kept_sites=self.add_sites,
            multi_allelic_sites=self.white_list,
        )

        het_sites = self.het_sites
        if self.del2_reads_partial != set():
            raw_read_haps, het_sites = self.update_reads_for_deletions(
                raw_read_haps,
                het_sites,
                self.del2_3p_pos1,
                self.del2_5p_pos2,
                self.del2_reads_partial,
                "3",
                self.deletion2_name,
            )
        if self.del1_reads_partial != set():
            raw_read_haps, het_sites = self.update_reads_for_deletions(
                raw_read_haps,
                het_sites,
                self.del1_3p_pos1,
                self.del1_5p_pos2,
                self.del1_reads_partial,
                "4",
                self.deletion1_name,
            )
        self.het_sites = het_sites

        # assemble haplotypes
        (
            ass_haps,
            original_haps,
            hcn,
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            read_counts,
        ) = self.phase_haps(raw_read_haps)

        tmp = {}
        for i, hap in enumerate(ass_haps):
            hap_name = f"hap{i+1}"
            tmp.setdefault(hap, hap_name)
        final_haps = tmp
        # get haps that extend into tnxb
        ending_copies = [
            final_haps[a]
            for a in ass_haps
            if a[0] not in ["0", "x"] and a[-1] not in ["0", "x"]
        ]
        starting_copies = [
            final_haps[a] for a in ass_haps if a[0] == "0" and a[-1] == "0"
        ]
        single_copies = [
            final_haps[a] for a in ass_haps if a[0] == "0" and a[-1] not in ["0", "x"]
        ]

        haplotypes = None
        if final_haps != {}:
            haplotypes = self.output_variants_in_haplotypes(
                final_haps,
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
                known_del={
                    "4": self.deletion1_name,
                    "3": self.deletion2_name,
                },
            )

        # phase haplotypes into alleles
        (
            alleles,
            hap_links,
            _,
            _,
        ) = self.phase_alleles(
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            final_haps,
            reverse=self.is_reverse,
        )

        successful_phasing, alleles, two_cp_haplotypes = self.update_alleles(
            alleles,
            haplotypes,
            final_haps,
            single_copies,
            starting_copies,
            ending_copies,
            hcn,
        )

        # annotate haplotypes by checking the diff sites
        # output variants carried by each haplotype
        hap_variants = {}
        if haplotypes is not None:
            for hap, hap_info in haplotypes.items():
                hap_variants.setdefault(hap, [])
                for var in hap_info["variants"]:
                    if var in self.known_variants:
                        hap_variants[hap].append(self.known_variants[var])

        total_cn = len(ass_haps) + len(two_cp_haplotypes)
        if ass_haps == [] and self.het_sites == []:
            # homozygous, feed all reads to call variants
            total_cn = 2
        if total_cn < 2 or len(ending_copies) > 2:
            total_cn = None

        annotated_alleles = self.annotate_alleles(
            successful_phasing,
            alleles,
            hap_variants,
            ending_copies,
            ass_haps,
            two_cp_haplotypes,
        )

        self.close_handle()

        return self.GeneCall(
            total_cn,
            final_haps,
            two_cp_haplotypes,
            starting_copies,
            ending_copies,
            single_copies,
            successful_phasing,
            alleles,
            annotated_alleles,
            hap_variants,
            hap_links,
            hcn,
            original_haps,
            self.het_sites,
            uniquely_supporting_reads,
            self.het_no_phasing,
            self.homo_sites,
            haplotypes,
            nonuniquely_supporting_reads,
            raw_read_haps,
            self.mdepth,
            self.region_avg_depth._asdict(),
        )



# paraphase
# Author: Xiao Chen <xchen@pacificbiosciences.com>

import copy
from collections import namedtuple
from ..phaser import Phaser


class Ncf1Phaser(Phaser):
    new_fields = copy.deepcopy(Phaser.fields)
    new_fields.insert(4, "pseudo_reads")
    new_fields.insert(4, "gene_reads")
    new_fields.remove("alleles_final")
    new_fields.remove("hap_links")
    GeneCall = namedtuple(
        "GeneCall",
        new_fields,
        defaults=(None,) * len(new_fields),
    )

    def __init__(
        self, sample_id, outdir, genome_depth=None, genome_bam=None, sample_sex=None
    ):
        Phaser.__init__(self, sample_id, outdir, genome_depth, genome_bam, sample_sex)

    def set_parameter(self, config):
        super().set_parameter(config)
        self.pivot_var = config.get("pivot_var")

    def call(self):
        if self.check_coverage_before_analysis() is False:
            return self.GeneCall()
        pivot_site = self.pivot_site
        for pileupcolumn in self._bamh.pileup(
            self.nchr, pivot_site, pivot_site + 1, truncate=True
        ):
            bases = [
                a.upper() for a in pileupcolumn.get_query_sequences(add_indels=True)
            ]
        gene_reads = bases.count("G")
        pseudo_reads = bases.count("G-2NN")

        self.get_homopolymer()
        self.get_candidate_pos()
        self.het_sites = sorted(list(self.candidate_pos))
        self.remove_noisy_sites()

        raw_read_haps = self.get_haplotypes_from_reads(add_sites=self.add_sites)

        (
            ass_haps,
            original_haps,
            hcn,
            uniquely_supporting_reads,
            nonuniquely_supporting_reads,
            raw_read_haps,
            read_counts,
        ) = self.phase_haps(raw_read_haps)

        total_cn = len(ass_haps)
        tmp = {}
        for i, hap in enumerate(ass_haps):
            tmp.setdefault(hap, f"hap{i+1}")
        ass_haps = tmp

        haplotypes = None
        if self.het_sites != []:
            haplotypes = self.output_variants_in_haplotypes(
                ass_haps,
                uniquely_supporting_reads,
                nonuniquely_supporting_reads,
            )

        hap_rename = {}
        counter_gene = 0
        counter_pseudo = 0
        # main variant is 74777266_GGT_G
        var_reads = self.check_variants_in_haplotypes(self.pivot_var)
        for hap_seq, hap in ass_haps.items():
            var = haplotypes[hap]["variants"]
            if self.pivot_var not in var:
                hap_reads = uniquely_supporting_reads[hap_seq]
                hap_var = [var_reads.get(a) for a in hap_reads]
                if hap_var.count("alt") > (len(hap_var) - hap_var.count(None)) * 0.7:
                    counter_pseudo += 1
                    hap_rename.setdefault(hap, f"pseudo_hap{counter_pseudo}")
                else:
                    counter_gene += 1
                    hap_rename.setdefault(hap, f"ncf1_hap{counter_gene}")
            else:
                counter_pseudo += 1
                hap_rename.setdefault(hap, f"pseudo_hap{counter_pseudo}")

        tmp = {}
        for hap, hap_name in ass_haps.items():
            tmp.setdefault(hap, hap_rename[hap_name])
        ass_haps = tmp

        tmp = {}
        for hap_name, hap_info in haplotypes.items():
            tmp.setdefault(hap_rename[hap_name], hap_info)
        haplotypes = tmp

        two_cp_haps = []
        if counter_gene == 1:
            two_cp_hap_candidate = self.compare_depth(haplotypes)
            if "ncf1_hap1" in two_cp_hap_candidate:
                two_cp_haps = two_cp_hap_candidate
                counter_gene += 1
                total_cn += 1

        if self.mdepth is not None:
            prob = self.depth_prob(gene_reads, self.mdepth / 2)
            if prob[0] < 0.9 and counter_gene == 1:
                counter_gene = None
                total_cn = None
            if prob[0] > 0.95 and counter_gene > 1 and two_cp_haps != []:
                counter_gene = None
                total_cn = None
        # scenario where only three haplotypes are found, possibly each at CN2
        if counter_gene == 1 and counter_pseudo == 2 and total_cn == 3:
            counter_gene = None
            total_cn = None

        # homozygous case
        if total_cn == 0:
            total_cn = None

        self.close_handle()

        return self.GeneCall(
            total_cn,
            counter_gene,
            ass_haps,
            two_cp_haps,
            gene_reads,
            pseudo_reads,
            hcn,
            original_haps,
            self.het_sites,
            uniquely_supporting_reads,
            self.het_no_phasing,
            self.homo_sites,
            haplotypes,
            nonuniquely_supporting_reads,
            raw_read_haps,
            self.mdepth,
            self.region_avg_depth._asdict(),
            self.sample_sex,
        )



import pytest
import yaml
import os
from paraphase.phaser import Phaser


def update_config(gene):
    cur_dir = os.path.dirname(__file__)
    sample_dir = os.path.join(cur_dir, "test_data")
    data_dir = os.path.join(os.path.dirname(cur_dir), "paraphase", "data", "38")
    config_file = os.path.join(data_dir, "config.yaml")
    with open(config_file, "r") as f:
        config = yaml.safe_load(f)
    config = config[gene]
    config.setdefault("gene", gene)
    realign_region = config["realign_region"]
    nchr = realign_region.split(":")[0]
    nchr_old = realign_region.replace(":", "_").replace("-", "_")
    config.setdefault("nchr", nchr)
    config.setdefault("nchr_old", nchr_old)
    if "data" in config:
        data_paths = config.get("data")
    else:
        data_paths = {}
    for data_entry in data_paths:
        old_data_file = data_paths[data_entry]
        new_data_file = os.path.join(data_dir, gene, old_data_file)
        data_paths[data_entry] = new_data_file
    data_paths.setdefault("reference", os.path.join(sample_dir, f"{gene}_ref.fa"))
    config.setdefault("data", data_paths)
    return config


class TestPhaser(object):
    sample_id = "HG00733"
    cur_dir = os.path.dirname(__file__)
    sample_dir = os.path.join(cur_dir, "test_data")
    phaser = Phaser(sample_id, sample_dir)
    config = update_config("smn1")
    phaser.set_parameter(config)

    def test_depth_prob(self):
        prob = Phaser.depth_prob(40, 20)
        assert prob.index(max(prob)) == 1

    def test_get_pivot_site_index(self):
        self.phaser.het_sites = ["70951940_A_C", "70951946_T_G"]
        pos, found_splice = self.phaser.get_pivot_site_index()
        assert pos == 1
        assert found_splice is True

        self.phaser.het_sites = ["70951940_A_C", "70951946_T_G", "70951958_T_G"]
        pos, found_splice = self.phaser.get_pivot_site_index()
        assert pos == 1
        assert found_splice is True

        self.phaser.het_sites = ["70951947_A_C", "70951949_T_G", "70951958_T_G"]
        pos, found_splice = self.phaser.get_pivot_site_index()
        assert pos == -1
        assert found_splice is False

    def test_process_indel(self):
        ref_seq, var_seq, indel_size = self.phaser.process_indel(
            70940935, "A", "A+3CCC"
        )
        assert ref_seq == "A"
        assert var_seq == "ACCC"
        assert indel_size == 3

        ref_seq, var_seq, indel_size = self.phaser.process_indel(70940935, "A", "A-2NN")
        assert ref_seq == "ACT"
        assert var_seq == "A"
        assert indel_size == 2

    def test_simplify_read_haps(self):
        haps = {"r1": ["1", "2"], "r2": ["2", "1"]}
        haplotypes_to_reads, reads_to_haplotypes = Phaser.simplify_read_haps(haps)
        assert haplotypes_to_reads == {"12": ["r1"], "21": ["r2"]}
        assert reads_to_haplotypes == {"r1": "12", "r2": "21"}

    def test_get_start_end(self):
        hap = "xx1211xxx"
        nstart, nend = Phaser.get_start_end(hap)
        assert nstart == 2
        assert nend == 5

    def test_get_hap_variant_ranges(self):
        self.phaser.het_sites = [
            "70917101_A_C",
            "70917111_A_C",
            "70917150_A_C",
            "70917200_A_C",
            "70917300_A_C",
            "70917400_A_C",
        ]
        hap = "x121xx"
        n1, n2 = self.phaser.get_hap_variant_ranges(hap)
        assert n1 == 70917102
        assert n2 == 70917299

        hap = "121211"
        n1, n2 = self.phaser.get_hap_variant_ranges(hap)
        assert n1 == 70917100
        assert n2 == 70961220

    def test_update_reads_for_deletions(self):
        self.phaser.het_sites = [
            "70917101_A_C",
            "70917111_A_C",
            "70917150_A_C",
            "70917200_A_C",
            "70917300_A_C",
            "70917400_A_C",
        ]
        del_reads_partial = {"r1"}
        raw_read_haps = {
            "r1": "1" * 6,
        }
        raw_read_haps, het_sites = self.phaser.update_reads_for_deletions(
            raw_read_haps,
            self.phaser.het_sites,
            70917199,
            70917201,
            del_reads_partial,
            "3",
            "del",
        )
        assert raw_read_haps["r1"] == "111311"
        assert het_sites == [
            "70917101_A_C",
            "70917111_A_C",
            "70917150_A_C",
            "70917200_A_C",
            "70917300_A_C",
            "70917400_A_C",
        ]

        # del at beginning, insert del at beginning
        self.phaser.het_sites = [
            "70917101_A_C",
            "70917111_A_C",
            "70917150_A_C",
            "70917200_A_C",
            "70917300_A_C",
            "70917400_A_C",
        ]
        del_reads_partial = {"r1"}
        raw_read_haps = {
            "r1": "1" * 6,
        }
        raw_read_haps, het_sites = self.phaser.update_reads_for_deletions(
            raw_read_haps,
            self.phaser.het_sites,
            70917100,
            70917100,
            del_reads_partial,
            "3",
            "del",
        )
        assert raw_read_haps["r1"] == "3111111"
        assert het_sites == [
            "del",
            "70917101_A_C",
            "70917111_A_C",
            "70917150_A_C",
            "70917200_A_C",
            "70917300_A_C",
            "70917400_A_C",
        ]

        # del at the end, do nothing?
        self.phaser.het_sites = [
            "70917101_A_C",
            "70917111_A_C",
            "70917150_A_C",
            "70917200_A_C",
            "70917300_A_C",
            "70917400_A_C",
        ]
        del_reads_partial = {"r1"}
        raw_read_haps = {
            "r1": "1" * 6,
        }
        raw_read_haps, het_sites = self.phaser.update_reads_for_deletions(
            raw_read_haps,
            self.phaser.het_sites,
            70917500,
            70917600,
            del_reads_partial,
            "3",
            "del",
        )
        assert raw_read_haps["r1"] == "111111"
        assert het_sites == [
            "70917101_A_C",
            "70917111_A_C",
            "70917150_A_C",
            "70917200_A_C",
            "70917300_A_C",
            "70917400_A_C",
        ]

    def test_check_linking_read(self):
        aln1 = "xxx12"
        aln2 = "11xxxx"
        assert self.phaser.check_linking_read(aln1, aln2) == "1-2"
        assert self.phaser.check_linking_read(aln1, aln2, reverse=True) is None

        aln1 = "11xxxx"
        aln2 = "xxx12"
        assert self.phaser.check_linking_read(aln1, aln2) == "2-1"
        assert self.phaser.check_linking_read(aln1, aln2, reverse=True) is None

        aln1 = "xxx12"
        aln2 = "xxxxxx11"
        assert self.phaser.check_linking_read(aln1, aln2) is None
        assert self.phaser.check_linking_read(aln1, aln2, reverse=True) == "0-0"

        aln1 = "xxx1x"
        aln2 = "1xxxxx11"
        assert self.phaser.check_linking_read(aln1, aln2) is None
        assert self.phaser.check_linking_read(aln1, aln2, reverse=True) is None

    def test_get_alleles_from_links(self):
        ass_haps = ["1", "2", "3", "4", "5", "6", "7"]
        read_links = {"1": ["2"], "2": ["1", "3"], "3": ["2"]}
        alleles = self.phaser.get_alleles_from_links(read_links, ass_haps)
        assert alleles == [["1", "2", "3"]]

        # test merge
        read_links = {
            "1": ["2"],
            "2": ["1", "3"],
            "4": ["5"],
            "5": ["4", "3"],
            "3": ["5"],
            "6": ["7"],
            "7": ["6"],
        }
        alleles = self.phaser.get_alleles_from_links(read_links, ass_haps)
        assert len(alleles) == 2 and (
            sorted(alleles[0]) == ["1", "2", "3", "4", "5"]
            or sorted(alleles[1]) == ["1", "2", "3", "4", "5"]
        )

    def test_get_directed_links(self):
        raw_read_haps = {
            "r1": "xx1",
            "r1_sup": "2xx",
            "r2": "xx1",
            "r2_sup": "2xx",
            "r3": "x1x",  # does not cover either end
            "r3_sup": "2xx",
        }
        new_reads = {
            "r1": [{"r1": ["111"]}, {"r1_sup": ["211"]}],
            "r2": [{"r2": ["111"]}, {"r2_sup": ["211"]}],
            "r3": [{"r3": ["111"]}, {"r3_sup": ["211"]}],
        }
        ass_haps = {"111": "hap1", "211": "hap2"}

        (
            nondirected_links,
            directed_links,
            directed_links_loose,
        ) = self.phaser.get_directed_links(new_reads, raw_read_haps, ass_haps, False)
        assert directed_links == {"hap1-hap2": [1, 1]}
        assert nondirected_links == {"hap1-hap2": [1, 1, 1]}
